# Copyright (C) 2018 NuCypher
#
# This file is part of nufhe.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

import numpy

from reikna.core import Computation, Transformation, Parameter, Annotation, Type
from reikna.algorithms import PureParallel, Reduce, predicate_sum
from reikna.cluda import functions, Snippet
import reikna.helpers as helpers
from reikna import transformations

from .numeric_functions import Torus32, Int32, ErrorFloat


TEMPLATE = helpers.template_for(__file__)


class MatrixMulVector(Computation):

    def __init__(self, matrix_t):
        Computation.__init__(self, [
            Parameter('output', Annotation(Type(matrix_t.dtype, matrix_t.shape[:-1]), 'o')),
            Parameter('matrix', Annotation(matrix_t, 'i')),
            Parameter('vector', Annotation(Type(matrix_t.dtype, matrix_t.shape[-1]), 'i'))])

    def _build_plan(self, plan_factory, device_params, output, matrix, vector):
        plan = plan_factory()

        summation = Reduce(matrix, predicate_sum(matrix.dtype), axes=(len(matrix.shape)-1,))

        mul_vec = Transformation([
            Parameter('output', Annotation(matrix, 'o')),
            Parameter('matrix', Annotation(matrix, 'i')),
            Parameter('vector', Annotation(vector, 'i'))],
            """
            ${output.store_same}(${mul}(${matrix.load_same}, ${vector.load_idx}(${idxs[-1]})));
            """,
            render_kwds=dict(mul=functions.mul(matrix.dtype, vector.dtype)),
            connectors=['output', 'matrix'])

        summation.parameter.input.connect(
            mul_vec, mul_vec.output, matrix=mul_vec.matrix, vector=mul_vec.vector)

        plan.computation_call(summation, output, matrix, vector)

        return plan


class MakeLweKeyswitchKey(Computation):

    def __init__(
            self, input_size: int, output_size: int,
            decomp_length: int, log2_base: int, noise: float):

        base = 2**log2_base

        a = Type(Torus32, (input_size, decomp_length, base, output_size))
        b = Type(Torus32, (input_size, decomp_length, base))
        cv = Type(ErrorFloat, (input_size, decomp_length, base))
        in_key = Type(Int32, input_size)
        out_key = Type(Int32, output_size)

        noises_a = Type(Torus32, (input_size, decomp_length, base - 1, output_size))
        noises_b = Type(Torus32, (input_size, decomp_length, base - 1))

        self._output_size = output_size
        self._log2_base = log2_base
        self._noise = noise

        Computation.__init__(self, [
            Parameter('ks_a', Annotation(a, 'o')),
            Parameter('ks_b', Annotation(b, 'o')),
            Parameter('ks_cv', Annotation(cv, 'o')),
            Parameter('in_key', Annotation(in_key, 'i')),
            Parameter('out_key', Annotation(out_key, 'i')),
            Parameter('noises_a', Annotation(noises_a, 'i')),
            Parameter('noises_b', Annotation(noises_b, 'i'))])

    def _build_plan(
            self, plan_factory, device_params,
            ks_a, ks_b, ks_cv, in_key, out_key, noises_a, noises_b):

        plan = plan_factory()

        extracted_n, t, base, inner_n = ks_a.shape

        mul_key = MatrixMulVector(noises_a)
        b_term = plan.temp_array_like(mul_key.parameter.output)

        build_keyswitch = PureParallel([
            Parameter('ks_a', Annotation(ks_a, 'o')),
            Parameter('ks_b', Annotation(ks_b, 'o')),
            Parameter('ks_cv', Annotation(ks_cv, 'o')),
            Parameter('in_key', Annotation(in_key, 'i')),
            Parameter('b_term', Annotation(b_term, 'i')),
            Parameter('noises_a', Annotation(noises_a, 'i')),
            Parameter('noises_b', Annotation(noises_b, 'i'))],
            Snippet(
                TEMPLATE.get_def("make_lwe_keyswitch_key"),
                render_kwds=dict(
                    log2_base=self._log2_base, output_size=self._output_size,
                    noise=self._noise)),
            guiding_array="ks_b")

        plan.computation_call(mul_key, b_term, noises_a, out_key)
        plan.computation_call(
            build_keyswitch,
            ks_a, ks_b, ks_cv, in_key, b_term, noises_a, noises_b)

        return plan


class LweKeyswitch(Computation):

    def __init__(
            self, result_shape_info,
            input_size: int, output_size: int, decomp_length: int, log2_base: int):

        base = 2**log2_base

        a = result_shape_info.a
        b = result_shape_info.b
        cv = result_shape_info.current_variances

        ks_a = Type(Torus32, (input_size, decomp_length, base, output_size))
        ks_b = Type(Torus32, (input_size, decomp_length, base))
        ks_cv = Type(ErrorFloat, (input_size, decomp_length, base))

        source_a = Type(Torus32, result_shape_info.shape + (input_size,))
        source_b = Type(Torus32, result_shape_info.shape)

        self._decomp_length = decomp_length
        self._input_size = input_size
        self._output_size = output_size
        self._log2_base = log2_base

        Computation.__init__(self, [
            Parameter('result_a', Annotation(a, 'io')),
            Parameter('result_b', Annotation(b, 'io')),
            Parameter('result_cv', Annotation(cv, 'io')),
            Parameter('ks_a', Annotation(ks_a, 'i')),
            Parameter('ks_b', Annotation(ks_b, 'i')),
            Parameter('ks_cv', Annotation(ks_cv, 'i')),
            Parameter('source_a', Annotation(source_a, 'i')),
            Parameter('source_b', Annotation(source_b, 'i'))])

    def _build_plan(
            self, plan_factory, device_params,
            result_a, result_b, result_cv,
            ks_a, ks_b, ks_cv, source_a, source_b):

        plan = plan_factory()

        batch_shape = result_a.shape[:-1]

        plan.kernel_call(
            TEMPLATE.get_def("lwe_keyswitch"),
            [result_a, result_b, result_cv, ks_a, ks_b, ks_cv, source_a, source_b],
            kernel_name="lwe_keyswitch",
            global_size=(helpers.product(batch_shape), self._output_size),
            render_kwds=dict(
                slices=(len(batch_shape), 1),
                output_size=self._output_size,
                input_size=self._input_size,
                log2_base=self._log2_base,
                decomp_length=self._decomp_length,
                ))

        return plan


class LweEncrypt(Computation):

    def __init__(self, shape, lwe_size: int, noise: float):

        result_a = Type(Torus32, shape + (lwe_size,))
        result_b = Type(Torus32, shape)
        result_cv = Type(ErrorFloat, shape)
        messages = Type(Torus32, shape)
        key = Type(Int32, (lwe_size,))
        noises_a = Type(Torus32, shape + (lwe_size,))
        noises_b = Type(Torus32, shape)

        self._noise = noise

        Computation.__init__(self, [
            Parameter('result_a', Annotation(result_a, 'o')),
            Parameter('result_b', Annotation(result_b, 'o')),
            Parameter('result_cv', Annotation(result_cv, 'o')),
            Parameter('messages', Annotation(messages, 'i')),
            Parameter('key', Annotation(key, 'i')),
            Parameter('noises_a', Annotation(noises_a, 'i')),
            Parameter('noises_b', Annotation(noises_b, 'i'))])

    def _build_plan(
            self, plan_factory, device_params,
            result_a, result_b, result_cv, messages, key, noises_a, noises_b):

        plan = plan_factory()

        mul_key = MatrixMulVector(noises_a)

        fill_b_cv = Transformation([
            Parameter('result_b', Annotation(result_b, 'o')),
            Parameter('result_cv', Annotation(result_cv, 'o')),
            Parameter('messages', Annotation(messages, 'i')),
            Parameter('noises_a_times_key', Annotation(noises_b, 'i')),
            Parameter('noises_b', Annotation(noises_b, 'i'))],
            """
            ${result_b.store_same}(
                ${noises_b.load_same}
                + ${messages.load_same}
                + ${noises_a_times_key.load_same});
            ${result_cv.store_same}(${noise**2});
            """,
            connectors=['noises_a_times_key'],
            render_kwds=dict(noise=self._noise))

        mul_key.parameter.output.connect(
            fill_b_cv, fill_b_cv.noises_a_times_key,
            b=fill_b_cv.result_b, cv=fill_b_cv.result_cv, messages=fill_b_cv.messages,
            noises_b=fill_b_cv.noises_b)

        plan.computation_call(mul_key, result_b, result_cv, messages, noises_b, noises_a, key)
        plan.computation_call(
            PureParallel.from_trf(transformations.copy(noises_a)),
            result_a, noises_a)

        return plan


class LweDecrypt(Computation):
    """
    Compute the phase of the sample using the secret key: phi = b - a.s
    """

    def __init__(self, shape, lwe_size):

        a = Type(Torus32, shape + (lwe_size,))
        b = Type(Torus32, shape)
        key = Type(Int32, (lwe_size,))

        Computation.__init__(self, [
            Parameter('result', Annotation(b, 'o')),
            Parameter('lwe_a', Annotation(a, 'i')),
            Parameter('lwe_b', Annotation(b, 'i')),
            Parameter('key', Annotation(key, 'i'))])

    def _build_plan(self, plan_factory, device_params, result, lwe_a, lwe_b, key):

        plan = plan_factory()

        mul_key = MatrixMulVector(lwe_a)

        fill_res = Transformation([
            Parameter('result', Annotation(result, 'o')),
            Parameter('b', Annotation(lwe_b, 'i')),
            Parameter('a_times_key', Annotation(lwe_b, 'i'))],
            """
            ${result.store_same}(${b.load_same} - ${a_times_key.load_same});
            """,
            connectors=['a_times_key'])

        mul_key.parameter.output.connect(
            fill_res, fill_res.a_times_key,
            result=fill_res.result, b=fill_res.b)

        plan.computation_call(mul_key, result, lwe_b, lwe_a, key)

        return plan


class LweLinear(Computation):

    def __init__(self, result_shape_info, source_shape_info, add_result=False):

        self._add_result = add_result

        Computation.__init__(self, [
            Parameter('result_a', Annotation(result_shape_info.a, 'o')),
            Parameter('result_b', Annotation(result_shape_info.b, 'o')),
            Parameter('result_cv', Annotation(result_shape_info.current_variances, 'o')),
            Parameter('source_a', Annotation(source_shape_info.a, 'i')),
            Parameter('source_b', Annotation(source_shape_info.b, 'i')),
            Parameter('source_cv', Annotation(source_shape_info.current_variances, 'i')),
            Parameter('coeff', Annotation(Type(Torus32)))])

    def _build_plan(
            self, plan_factory, device_params,
            result_a, result_b, result_cv, source_a, source_b, source_cv, coeff):

        plan = plan_factory()
        plan.kernel_call(
            TEMPLATE.get_def("lwe_linear"),
            [result_a, result_b, result_cv, source_a, source_b, source_cv, coeff],
            kernel_name="lwe_linear",
            global_size=result_a.shape,
            render_kwds=dict(
                add_result=self._add_result,
                ))

        return plan


class LweNoiselessTrivial(Computation):

    def __init__(self, result_shape_info, source_shape):
        Computation.__init__(self, [
            Parameter('result_a', Annotation(result_shape_info.a, 'o')),
            Parameter('result_b', Annotation(result_shape_info.b, 'o')),
            Parameter('result_cv', Annotation(result_shape_info.current_variances, 'o')),
            Parameter('mus', Annotation(Type(Torus32, source_shape), 'i'))])

    def _build_plan(self, plan_factory, device_params, result_a, result_b, result_cv, mus):

        plan = plan_factory()
        plan.kernel_call(
            TEMPLATE.get_def("lwe_noiseless_trivial"),
            [result_a, result_b, result_cv, mus],
            kernel_name="lwe_noiseless_trivial",
            global_size=result_a.shape)

        return plan


def LweNoiselessTrivialConstant(shape_info):
    comp = LweNoiselessTrivial(shape_info, shape_info.shape)
    bc = transformations.broadcast_param(comp.parameter.mus)
    comp.parameter.mus.connect(bc, bc.output, mu=bc.param)
    return comp



# Copyright (C) 2018 NuCypher
#
# This file is part of nufhe.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

import pickle

from .numeric_functions import Torus32, t32_to_phase
from .polynomials import TorusPolynomialArray, shift_tp_inverted_power
from .lwe import LweParams, LweKey, LweSampleArray, LweKeyswitchKey, lwe_keyswitch
from .tgsw import (
    TGswKey,
    TransformedTGswSampleArray,
    TGswParams,
    TGswSampleArray,
    tgsw_transform_samples,
    tgsw_encrypt_int,
    tgsw_transformed_external_mul,
    )
from .tlwe import (
    TLweSampleArray,
    tlwe_noiseless_trivial,
    tlwe_shift_polynomials,
    tlwe_add_to,
    tlwe_extract_lwe_samples,
    tlwe_copy,
    )
from .blind_rotate import BlindRotate_gpu
from .performance import PerformanceParametersForDevice


class BootstrapKey:

    def __init__(
            self, in_out_params: LweParams, tgsw: TransformedTGswSampleArray):

        bk_params = tgsw.params
        accum_params = bk_params.tlwe_params
        extract_params = accum_params.extracted_lweparams

        self.in_out_params = in_out_params
        self.bk_params = bk_params
        self.accum_params = accum_params
        self.extract_params = extract_params
        self.tgsw = tgsw

    @classmethod
    def from_rng(
            cls, thr, rng, lwe_key: LweKey, tgsw_key: TGswKey,
            perf_params: PerformanceParametersForDevice):

        in_out_params = lwe_key.params
        bk_params = tgsw_key.params
        accum_params = bk_params.tlwe_params

        # Make a non-transformed bootstrap key
        bk = TGswSampleArray.empty(thr, bk_params, (in_out_params.size,))
        tgsw_encrypt_int(thr, rng, bk, lwe_key.key, accum_params.min_noise, tgsw_key, perf_params)

        # Convert it to transformed space, because that's where it will be used
        bk_transformed = TransformedTGswSampleArray.empty(thr, bk_params, (in_out_params.size,))
        tgsw_transform_samples(thr, bk_transformed, bk, perf_params)

        return cls(in_out_params, bk_transformed)

    def dump(self, file_obj):
        pickle.dump(self.in_out_params, file_obj)
        self.tgsw.dump(file_obj)

    @classmethod
    def load(cls, file_obj, thr):
        in_out_params = pickle.load(file_obj)
        tgsw = TransformedTGswSampleArray.load(file_obj, thr)
        return cls(in_out_params, tgsw)

    def __eq__(self, other: 'BootstrapKey'):
        return (
            self.__class__ == other.__class__
            and self.in_out_params == other.in_out_params
            and self.tgsw == other.tgsw)



def mux_rotate(
        thr, result: TLweSampleArray, accum: TLweSampleArray, bki: TransformedTGswSampleArray,
        bk_idx: int, barai, bk_params: TGswParams, perf_params: PerformanceParametersForDevice):

    # TYPING: barai::Array{Int32}
    # ACC = BKi*[(X^barai-1)*ACC]+ACC
    # temp = (X^barai-1)*ACC
    tlwe_shift_polynomials(thr, result, accum, barai, bk_idx)

    # temp *= BKi
    tgsw_transformed_external_mul(thr, result, bki, bk_idx, perf_params)

    # ACC += temp
    tlwe_add_to(thr, result, accum)


"""
 * multiply the accumulator by X^sum(bara_i.s_i)
 * @param accum the TLWE sample to multiply
 * @param bk An array of n TGSW FFT samples where bk_i encodes s_i
 * @param bara An array of n coefficients between 0 and 2N-1
 * @param bk_params The parameters of bk
"""
def blind_rotate(
        thr, accum: TLweSampleArray, bk: BootstrapKey,
        bara, n: int, bk_params: TGswParams, perf_params: PerformanceParametersForDevice):

    # TYPING: bara::Array{Int32}
    temp = TLweSampleArray.empty(thr, bk_params.tlwe_params, accum.shape)

    temp2 = temp
    temp3 = accum

    accum_in_temp3 = True

    for i in range(n):
        # TODO: here we only need to pass bootstrap_key[i] and bara[:,i],
        # but Reikna kernels have to be recompiled for every set of strides/offsets,
        # so for now we are just passing full arrays and an index.
        mux_rotate(thr, temp2, temp3, bk.tgsw, i, bara, bk_params, perf_params)

        temp2, temp3 = temp3, temp2
        accum_in_temp3 = not accum_in_temp3

    # TODO: add a test that checks this
    if not accum_in_temp3: # temp3 != accum
        tlwe_copy(thr, accum, temp3)


"""
 * result = LWE(v_p) where p=barb-sum(bara_i.s_i) mod 2N
 * @param result the output LWE sample
 * @param v a 2N-elt anticyclic function (represented by a TorusPolynomial)
 * @param bk An array of n TGSW FFT samples where bk_i encodes s_i
 * @param barb A coefficients between 0 and 2N-1
 * @param bara An array of n coefficients between 0 and 2N-1
 * @param bk_params The parameters of bk
"""
def blind_rotate_and_extract(
        thr, result: LweSampleArray,
        v: TorusPolynomialArray, bk: BootstrapKey, ks: LweKeyswitchKey,
        barb, bara,
        perf_params: PerformanceParametersForDevice,
        no_keyswitch=False):

    # TYPING: barb::Array{Int32},
    # TYPING: bara::Array{Int32}

    bk_params = bk.bk_params

    if not no_keyswitch:
        extracted_result = LweSampleArray.empty(
            thr, bk.accum_params.extracted_lweparams, result.shape_info.shape)
    else:
        extracted_result = result

    accum_params = bk_params.tlwe_params
    extract_params = accum_params.extracted_lweparams
    N = accum_params.polynomial_degree

    # testvector = X^{2N-barb}*v
    testvectbis = TorusPolynomialArray.empty(thr, N, extracted_result.shape_info.shape)
    shift_tp_inverted_power(thr, testvectbis, barb, v)

    # Accumulator
    acc = TLweSampleArray.empty(thr, accum_params, extracted_result.shape_info.shape)
    tlwe_noiseless_trivial(thr, acc, testvectbis)

    if perf_params.single_kernel_bootstrap:
        # includes blindrotate, extractlwesample and (optionally) keyswitch
        BlindRotate_gpu(result, acc, bk, ks, bara, perf_params, no_keyswitch=no_keyswitch)
    else:
        # Blind rotation
        blind_rotate(
            thr, acc, bk, bara, bk.in_out_params.size, bk_params, perf_params)

        # Extraction
        tlwe_extract_lwe_samples(thr, extracted_result, acc)

        if not no_keyswitch:
            lwe_keyswitch(thr, result, ks, extracted_result)


"""
 * result = LWE(mu) iff phase(x)>0, LWE(-mu) iff phase(x)<0
 * @param result The resulting LweSample
 * @param bk The bootstrapping + keyswitch key
 * @param mu The output message (if phase(x)>0)
 * @param x The input sample
"""
def bootstrap(
        thr, result: LweSampleArray, bk: BootstrapKey, ks: LweKeyswitchKey,
        mu: Torus32, x: LweSampleArray, perf_params: PerformanceParametersForDevice,
        no_keyswitch=False):

    accum_params = bk.accum_params
    N = accum_params.polynomial_degree

    testvect = TorusPolynomialArray.empty(thr, N, result.shape_info.shape)

    # Modulus switching
    barb = thr.array(x.b.shape, Torus32)
    bara = thr.array(x.a.shape, Torus32)

    t32_to_phase(thr, barb, x.b, 2 * N)
    t32_to_phase(thr, bara, x.a, 2 * N)

    # the initial testvec = [mu,mu,mu,...,mu]
    testvect.coeffs.fill(mu)

    # Bootstrapping rotation and extraction
    blind_rotate_and_extract(
        thr, result, testvect, bk, ks, barb, bara, perf_params,
        no_keyswitch=no_keyswitch)



import random
import nufhe

size = 32
bits1 = [random.choice([False, True]) for i in range(size)]
bits2 = [random.choice([False, True]) for i in range(size)]
reference = [not (b1 and b2) for b1, b2 in zip(bits1, bits2)]

ctx = nufhe.Context()
secret_key, cloud_key = ctx.make_key_pair()

ciphertext1 = ctx.encrypt(secret_key, bits1)
ciphertext2 = ctx.encrypt(secret_key, bits2)

vm = ctx.make_virtual_machine(cloud_key)
result = vm.gate_nand(ciphertext1, ciphertext2)
result_bits = ctx.decrypt(secret_key, result)

assert all(result_bits == reference)



# Copyright (C) 2018 NuCypher
#
# This file is part of nufhe.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

import numpy

from reikna.algorithms import PureParallel
from reikna.core import Computation, Transformation, Parameter, Annotation, Type
from reikna.cluda import dtypes

from .computation_cache import get_computation


# Declaring the types here instead of `numeric_functions.py` to avoid a cricular import.

# Element on a torus
Torus32 = numpy.int32

# Accompanying integer type (same size)
Int32 = numpy.int32

# The type for secondary floating-point values (e.g., errors)
ErrorFloat = numpy.float32


class Torus32ToPhase(Computation):

    def __init__(self, shape, mspace_size):

        self._mspace_size = mspace_size

        messages = Type(Torus32, shape)
        result = Type(Int32, shape)

        Computation.__init__(self, [
            Parameter('result', Annotation(result, 'o')),
            Parameter('messages', Annotation(messages, 'i'))])

    def _build_plan(self, plan_factory, device_params, result, phase):
        plan = plan_factory()

        tr = Transformation(
            [
                Parameter('result', Annotation(result, 'o')),
                Parameter('phase', Annotation(phase, 'i')),
            ],
            """
            <%
                interv = 2**32 // mspace_size
                half_interv = interv // 2
            %>
            ${phase.ctype} phase = ${phase.load_same};
            ${result.store_same}(((unsigned int)phase + ${half_interv}) / ${interv});
            """,
            render_kwds=dict(
                mspace_size=self._mspace_size,
                uint64=dtypes.ctype(numpy.uint64)),
            connectors=['result', 'phase'])

        plan.computation_call(
            PureParallel.from_trf(tr, guiding_array='result'),
            result, phase)

        return plan



# Copyright (C) 2018 NuCypher
#
# This file is part of nufhe.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

import numpy

from reikna.core import Computation, Transformation, Parameter, Annotation, Type
from reikna.algorithms import PureParallel
import reikna.helpers as helpers

from .numeric_functions import Torus32, Int32, ErrorFloat
from .polynomial_transform import get_transform
from .performance import PerformanceParametersForDevice


TEMPLATE = helpers.template_for(__file__)


class TLweNoiselessTrivial(Computation):

    def __init__(self, params: 'TLweParams', shape):
        a_type = Type(Torus32, shape + (params.mask_size + 1, params.polynomial_degree))
        cv_type = Type(ErrorFloat, shape)
        mu_type = Type(Torus32, shape + (params.polynomial_degree,))

        self._mask_size = params.mask_size

        Computation.__init__(self,
            [Parameter('a', Annotation(a_type, 'o')),
            Parameter('current_variances', Annotation(cv_type, 'o')),
            Parameter('mu', Annotation(mu_type, 'i'))])

    def _build_plan(self, plan_factory, device_params, a, current_variances, mu):
        plan = plan_factory()

        fill = PureParallel([
            Parameter('a', Annotation(a, 'o')),
            Parameter('current_variances', Annotation(current_variances, 'o')),
            Parameter('mu', Annotation(mu, 'i'))],
            """
            ${a.ctype} a;
            if (${idxs[-2]} == ${mask_size})
            {
                a = ${mu.load_idx}(${", ".join(idxs[:-2])}, ${idxs[-1]});
            }
            else
            {
                a = 0;
            }
            ${a.store_same}(a);

            if (${idxs[-1]} == 0 && ${idxs[-2]} == 0)
            {
                ${current_variances.store_idx}(${", ".join(idxs[:-2])}, 0);
            }
            """,
            render_kwds=dict(mask_size=self._mask_size))

        plan.computation_call(fill, a, current_variances, mu)

        return plan


class TLweExtractLweSamples(Computation):

    def __init__(self, params: 'TLweParams', shape):

        self._mask_size = params.mask_size
        self._polynomial_degree = params.polynomial_degree

        result_a = Type(Torus32, shape + (params.extracted_lweparams.size,))
        result_b = Type(Torus32, shape)
        tlwe_a = Type(Torus32, shape + (params.mask_size + 1, params.polynomial_degree))

        Computation.__init__(self, [
            Parameter('result_a', Annotation(result_a, 'o')),
            Parameter('result_b', Annotation(result_b, 'o')),
            Parameter('tlwe_a', Annotation(tlwe_a, 'i'))])

    def _build_plan(self, plan_factory, device_params, result_a, result_b, tlwe_a):
        plan = plan_factory()

        batch_len = helpers.product(result_b.shape)

        plan.kernel_call(
            TEMPLATE.get_def('tlwe_extract_lwe_samples'),
            [result_a, result_b, tlwe_a],
            kernel_name="tlwe_extract_lwe_samples",
            global_size=(batch_len, self._mask_size, self._polynomial_degree),
            render_kwds=dict(
                slices=(len(result_b.shape), 1, 1),
                mask_size=self._mask_size,
                polynomial_degree=self._polynomial_degree))

        return plan


class TLweEncryptZero(Computation):

    def __init__(
            self, params: 'TLweParams', shape, noise: float,
            perf_params: PerformanceParametersForDevice):

        polynomial_degree = params.polynomial_degree
        mask_size = params.mask_size

        result_a = Type(Torus32, shape + (mask_size + 1, polynomial_degree))
        result_cv = Type(ErrorFloat, shape)
        key = Type(Int32, (mask_size, polynomial_degree))
        noises1 = Type(Torus32, shape + (mask_size, polynomial_degree))
        noises2 = Type(Torus32, shape + (polynomial_degree,))

        self._transform_type = params.transform_type
        self._noise = noise
        self._mask_size = mask_size
        self._polynomial_degree = polynomial_degree
        self._perf_params = perf_params

        Computation.__init__(self, [
            Parameter('result_a', Annotation(result_a, 'o')),
            Parameter('result_cv', Annotation(result_cv, 'o')),
            Parameter('key', Annotation(key, 'i')),
            Parameter('noises1', Annotation(noises1, 'i')),
            Parameter('noises2', Annotation(noises2, 'i'))])

    def _build_plan(
            self, plan_factory, device_params,
            result_a, result_cv, key, noises1, noises2):

        plan = plan_factory()

        polynomial_degree = self._polynomial_degree
        batch_shape = result_a.shape[:-2]
        batch_len = helpers.product(batch_shape)

        perf_params = self._perf_params

        transform = get_transform(self._transform_type)

        ft_key = transform.ForwardTransform(key.shape[:-1], polynomial_degree, perf_params)
        key_tr = plan.temp_array_like(ft_key.parameter.output)

        ft_noises = transform.ForwardTransform(noises1.shape[:-1], polynomial_degree, perf_params)
        noises1_tr = plan.temp_array_like(ft_noises.parameter.output)

        ift = transform.InverseTransform(noises1.shape[:-1], polynomial_degree, perf_params)
        ift_res = plan.temp_array_like(ift.parameter.output)

        mul_tr = Transformation(
            [
                Parameter('output', Annotation(ift.parameter.input, 'o')),
                Parameter('key', Annotation(key_tr, 'i')),
                Parameter('noises1', Annotation(noises1_tr, 'i'))
            ],
            """
            ${output.store_same}(${tr_ctype}unpack(${mul}(
                ${tr_ctype}pack(${key.load_idx}(${idxs[-2]}, ${idxs[-1]})),
                ${tr_ctype}pack(${noises1.load_same})
                )));
            """,
            connectors=['output', 'noises1'],
            render_kwds=dict(
                mul=transform.transformed_mul(perf_params),
                tr_ctype=transform.transformed_internal_ctype()))

        ift.parameter.input.connect(mul_tr, mul_tr.output, key=mul_tr.key, noises1=mul_tr.noises1)

        plan.computation_call(ft_key, key_tr, key)
        plan.computation_call(ft_noises, noises1_tr, noises1)
        plan.computation_call(ift, ift_res, key_tr, noises1_tr)
        plan.kernel_call(
            TEMPLATE.get_def("tlwe_encrypt_zero_fill_result"),
            [result_a, result_cv, noises1, noises2, ift_res],
            kernel_name="tlwe_encrypt_zero_fill_result",
            global_size=(batch_len, self._mask_size + 1, polynomial_degree),
            render_kwds=dict(
                noise=self._noise, mask_size=self._mask_size,
                noises1_slices=(len(batch_shape), 1, 1),
                noises2_slices=(len(batch_shape), 1),
                cv_slices=(len(batch_shape),)
                ))

        return plan


class TLweTransformSamples(Computation):
    """
    Convert given Torus32 values to the transformed space and prepare them for fast multiplication
    (converting to Montgomery for NTT and doing nothing for FFT).
    """

    def __init__(
            self, params: 'TLweParams', shape, perf_params: PerformanceParametersForDevice):

        self._transform_type = params.transform_type
        self._perf_params = perf_params

        batch_shape = shape[:-1]
        polynomial_degree = params.polynomial_degree
        transform = get_transform(self._transform_type)
        tlength = transform.transformed_length(polynomial_degree)
        tdtype = transform.transformed_dtype()

        prepared_values = Type(tdtype, batch_shape + (tlength,))
        values = Type(Torus32, shape)

        Computation.__init__(self, [
            Parameter('prepared_samples', Annotation(prepared_values, 'o')),
            Parameter('values', Annotation(values, 'i'))])

    def _build_plan(self, plan_factory, device_params, prepared_values, values):

        plan = plan_factory()

        transform = get_transform(self._transform_type)

        comp = transform.ForwardTransform(values.shape[:-1], values.shape[-1], self._perf_params)
        prepare = transform.get_prepare_for_mul_trf(prepared_values.shape)
        comp.parameter.output.connect(prepare, prepare.input, prepared=prepare.output)

        plan.computation_call(comp, prepared_values, values)

        return plan



# Copyright (C) 2018 NuCypher
#
# This file is part of nufhe.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

import numpy

from reikna.helpers import product


# minus_one=True: result = (X^ai-1) * source
# minus_one=False: result = X^{a} * source
def ShiftTorusPolynomialReference(
        polynomial_degree, shape, powers_shape,
        powers_view=False, minus_one=False, invert_powers=False):

    batch_shape = powers_shape[:-1] if powers_view else powers_shape
    assert batch_shape == shape[:len(batch_shape)]
    poly_batch_shape = shape[len(batch_shape):]

    def _kernel(result, source, powers, powers_idx):

        if powers_view:
            powers = powers.reshape(product(batch_shape), powers_shape[-1])[:, powers_idx]
        else:
            powers = powers.flatten()

        result = result.reshape(product(batch_shape), product(poly_batch_shape), polynomial_degree)
        source = source.reshape(product(batch_shape), product(poly_batch_shape), polynomial_degree)

        if invert_powers:
            powers = 2 * polynomial_degree - powers

        for i in range(result.shape[0]):
            power = powers[i]
            if power < polynomial_degree:
                result[i,:,:power] = -source[i,:,(polynomial_degree - power):polynomial_degree]
                result[i,:,power:polynomial_degree] = source[i,:,:(polynomial_degree - power)]
            else:
                power = power - polynomial_degree
                result[i,:,:power] = source[i,:,(polynomial_degree - power):polynomial_degree]
                result[i,:,power:polynomial_degree] = -source[i,:,:(polynomial_degree - power)]

        if minus_one:
            result -= source

    return _kernel



# Copyright (C) 2018 NuCypher
#
# This file is part of nufhe.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

from .api_low_level import (
    make_key_pair,
    encrypt,
    decrypt,
    empty_ciphertext,
    NuFHEParameters,
    NuFHESecretKey,
    NuFHECloudKey,
    )

from .lwe import (
    LweSampleArray,
    concatenate,
    )

from .gates import (
    gate_nand,
    gate_or,
    gate_and,
    gate_xor,
    gate_xnor,
    gate_not,
    gate_copy,
    gate_constant,
    gate_nor,
    gate_andny,
    gate_andyn,
    gate_orny,
    gate_oryn,
    gate_mux,
    )

from .performance import PerformanceParameters

from .random_numbers import DeterministicRNG, SecureRNG

from .computation_cache import clear_computation_cache

from .api_high_level import (
    find_devices,
    Context,
    )



# Copyright (C) 2018 NuCypher
#
# This file is part of nufhe.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

import numpy

from reikna import helpers
from reikna.core import Computation, Transformation, Parameter, Annotation, Type
from reikna.algorithms import PureParallel
from reikna.cluda import dtypes

from .numeric_functions import Torus32, Int32


TEMPLATE = helpers.template_for(__file__)


class ShiftTorusPolynomial(Computation):
    """
    Calculate batched

        `result = (X^((2*N - p) if invert_powers else p) - minus_one) * source`

    where `N` is the polynomial degree.

    If `powers_view` is `True`, a view of `powers` is taken
    using fixed `powers_idx` as the innermost index.
    """

    def __init__(
            self, polynomial_degree, shape, powers_shape,
            powers_view=False, minus_one=False, invert_powers=False):

        self._batch_shape = powers_shape[:-1] if powers_view else powers_shape
        assert self._batch_shape == shape[:len(self._batch_shape)]

        self._powers_view = powers_view
        self._minus_one = minus_one
        self._invert_powers = invert_powers

        polynomials = Type(Torus32, shape + (polynomial_degree,))
        powers = Type(Int32, powers_shape)

        Computation.__init__(self, [
            Parameter('result', Annotation(polynomials, 'o')),
            Parameter('source', Annotation(polynomials, 'i')),
            Parameter('powers', Annotation(powers, 'i')),
            Parameter('powers_idx', Annotation(Type(Int32))) # unused if powers_view==False
            ])

    def _build_plan(self, plan_factory, device_params, result, source, powers, powers_idx):

        poly_batch_shape = result.shape[len(self._batch_shape):-1]
        polynomial_degree = result.shape[-1]

        plan = plan_factory()
        plan.kernel_call(
            TEMPLATE.get_def("shift_torus_polynomial"),
            [result, source, powers, powers_idx],
            kernel_name="shift_torus_polynomial",
            global_size=(
                helpers.product(self._batch_shape),
                helpers.product(poly_batch_shape),
                polynomial_degree),
            render_kwds=dict(
                batch_len=len(self._batch_shape),
                poly_batch_len=len(poly_batch_shape),
                polynomial_degree=polynomial_degree,
                powers_view=self._powers_view,
                minus_one=self._minus_one,
                invert_powers=self._invert_powers))

        return plan



# Copyright (C) 2018 NuCypher
#
# This file is part of nufhe.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

import random
import io

import numpy
import pytest

import nufhe


@pytest.mark.parametrize('make_pair', [False, True], ids=["make keys separately", "make_key_pair"])
def test_context(make_pair):

    size = 32
    bits1 = [random.choice([False, True]) for i in range(size)]
    bits2 = [random.choice([False, True]) for i in range(size)]
    reference = [not (b1 and b2) for b1, b2 in zip(bits1, bits2)]

    ctx = nufhe.Context()

    if make_pair:
        secret_key, cloud_key = ctx.make_key_pair()
    else:
        secret_key = ctx.make_secret_key()
        cloud_key = ctx.make_cloud_key(secret_key)

    ciphertext1 = ctx.encrypt(secret_key, bits1)
    ciphertext2 = ctx.encrypt(secret_key, bits2)

    vm = ctx.make_virtual_machine(cloud_key)
    result = vm.gate_nand(ciphertext1, ciphertext2)
    result_bits = ctx.decrypt(secret_key, result)

    assert all(result_bits == reference)


@pytest.fixture(scope='module', params=[False, True], ids=['to_file', 'to_string'])
def to_file(request):
    return request.param


def test_serialize_secret_key(to_file):

    ctx = nufhe.Context()
    secret_key = ctx.make_secret_key()

    if to_file:
        file_obj = io.BytesIO()
        secret_key.dump(file_obj)
        file_obj.seek(0)
        secret_key_loaded = ctx.load_secret_key(file_obj)
    else:
        s = secret_key.dumps()
        secret_key_loaded = ctx.load_secret_key(s)

    assert secret_key_loaded == secret_key


def test_serialize_cloud_key(to_file):

    ctx = nufhe.Context()
    secret_key, cloud_key = ctx.make_key_pair()

    if to_file:
        file_obj = io.BytesIO()
        cloud_key.dump(file_obj)
        file_obj.seek(0)
        cloud_key_loaded = ctx.load_cloud_key(file_obj)
    else:
        s = cloud_key.dumps()
        cloud_key_loaded = ctx.load_cloud_key(s)

    assert cloud_key_loaded == cloud_key


def test_serialize_ciphertext(to_file):

    ctx = nufhe.Context()
    secret_key, cloud_key = ctx.make_key_pair()

    size = 32
    bits = [random.choice([False, True]) for i in range(size)]
    ciphertext = ctx.encrypt(secret_key, bits)

    if to_file:
        file_obj = io.BytesIO()
        ciphertext.dump(file_obj)
        file_obj.seek(0)
        ciphertext_loaded = ctx.load_ciphertext(file_obj)
    else:
        s = ciphertext.dumps()
        ciphertext_loaded = ctx.load_ciphertext(s)

    assert ciphertext_loaded == ciphertext


@pytest.mark.parametrize('rng_cls', [nufhe.DeterministicRNG, nufhe.SecureRNG])
def test_rngs(rng_cls):

    size = 32
    bits1 = [random.choice([False, True]) for i in range(size)]
    bits2 = [random.choice([False, True]) for i in range(size)]
    reference = [not (b1 and b2) for b1, b2 in zip(bits1, bits2)]

    rng = rng_cls()
    ctx = nufhe.Context(rng=rng)
    secret_key, cloud_key = ctx.make_key_pair()

    ciphertext1 = ctx.encrypt(secret_key, bits1)
    ciphertext2 = ctx.encrypt(secret_key, bits2)

    vm = ctx.make_virtual_machine(cloud_key)
    result = vm.gate_nand(ciphertext1, ciphertext2)
    result_bits = ctx.decrypt(secret_key, result)

    assert all(result_bits == reference)


def test_vm_gate_constant(context_and_key_pair):
    """
    A regression for the result shape derivation in VM gate wrappers.
    The low-level ``gate_constant()`` accepts lists as arguments,
    but its VM wrapper did not.
    """

    ctx, secret_key, cloud_key = context_and_key_pair

    bits = [True, False, False, True]

    vm = ctx.make_virtual_machine(cloud_key)
    result = vm.gate_constant(bits)
    result_bits = ctx.decrypt(secret_key, result)

    assert (result_bits == bits).all()


def test_vm_gate_broadcasting(context_and_key_pair):
    """
    Check that VM gate wrapper chooses the correct shape of the resulting ciphertext to allocate,
    taking into account shapes of all the arguments.
    """

    ctx, secret_key, cloud_key = context_and_key_pair

    bits1 = numpy.random.randint(0, 2, size=(4,)).astype(numpy.bool)
    bits2 = numpy.random.randint(0, 2, size=(2, 4)).astype(numpy.bool)
    reference = numpy.logical_not(numpy.logical_and(bits1, bits2))

    ciphertext1 = ctx.encrypt(secret_key, bits1)
    ciphertext2 = ctx.encrypt(secret_key, bits2)

    vm = ctx.make_virtual_machine(cloud_key)
    result = vm.gate_nand(ciphertext1, ciphertext2)
    result_bits = ctx.decrypt(secret_key, result)

    assert (result_bits == reference).all()



# Copyright (C) 2018 NuCypher
#
# This file is part of nufhe.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

import numpy

from nufhe.transform.ntt_cpu import (
    GaloisNumber, gnum, ntt_naive, ntt, find_generator, gnum_to_i32)


def test_ntt():
    # Compare the naive NTT (O(N^2) complexity)
    # with the NTT based on the FFT scheme (O(N log N) complexity)

    a = gnum(numpy.random.randint(0, 1000, size=16))

    af = ntt(a, False)
    ab = ntt(af, True)

    af_ref = ntt_naive(a, False)
    ab_ref = ntt_naive(af_ref, True)

    assert (a == ab).all()
    assert (af == af_ref).all()
    assert (ab == ab_ref).all()


def test_find_generator():
    # Check that the returned generator

    g = find_generator(start=2)
    modulus = GaloisNumber.modulus

    for q in GaloisNumber.factors:
        assert g**((modulus - 1) // q) != 1

    assert g**(modulus - 1) == 1


def test_repr():
    # A serialization test
    a = GaloisNumber(1000)
    assert repr(a) == "GaloisNumber(1000)"
    assert str(a) == "1000G"


def test_gnum_to_i32():
    modulus = GaloisNumber.modulus
    assert gnum_to_i32(GaloisNumber(12345)) == 12345
    assert gnum_to_i32(GaloisNumber(2**31 + 12345)) == -2**31 + 12345
    assert gnum_to_i32(GaloisNumber(2**32 + 12345)) == 12345
    assert gnum_to_i32(GaloisNumber(modulus - 12345)) == -12345
    assert gnum_to_i32(GaloisNumber(modulus - (2**31 + 12345))) == 2**31 - 12345
    assert gnum_to_i32(GaloisNumber(modulus - (2**32 + 12345))) == -12345




# Copyright (C) 2018 NuCypher
#
# This file is part of nufhe.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

"""
Torus LWE functions.
"""

import pickle

import numpy

from .utils import arrays_equal
from .polynomial_transform import get_transform
from .computation_cache import get_computation
from .numeric_functions import ErrorFloat
from .random_numbers import rand_uniform_bool
from .lwe import LweParams, LweSampleArray
from .polynomials import (
    TorusPolynomialArray,
    IntPolynomialArray,
    TransformedPolynomialArray,
    shift_tp_minus_one_power_from_array,
    )
from .random_numbers import rand_gaussian_torus32, rand_uniform_torus32
from .performance import PerformanceParametersForDevice
from .tlwe_gpu import (
    TLweNoiselessTrivial,
    TLweExtractLweSamples,
    TLweEncryptZero,
    TLweTransformSamples,
    )


class TLweParams:

    def __init__(
            self, polynomial_degree: int, mask_size: int,
            min_noise: float, max_noise: float, transform_type):

        self.polynomial_degree = polynomial_degree # must be a power of 2
        self.mask_size = mask_size # number of polynomials in the mask
        self.min_noise = min_noise # minimum noise s.t. the sample is secure
        self.max_noise = max_noise # maximum noise s.t. we can decrypt
        self.extracted_lweparams = LweParams(
            polynomial_degree * mask_size, min_noise, max_noise) # LWE params if one extracts
        self.transform_type = transform_type

    def __eq__(self, other: 'TLweParams'):
        return (
            self.__class__ == other.__class__
            and self.polynomial_degree == other.polynomial_degree
            and self.mask_size == other.mask_size
            and self.min_noise == other.min_noise
            and self.max_noise == other.max_noise
            and self.transform_type == other.transform_type)

    def __hash__(self):
        return hash((
            self.__class__, self.polynomial_degree, self.mask_size,
            self.min_noise, self.max_noise, self.transform_type))


class TLweKey:

    def __init__(self, params: TLweParams, key):
        self.params = params
        self.key = key

    @classmethod
    def from_rng(cls, thr, params: TLweParams, rng):
        polynomial_degree = params.polynomial_degree
        mask_size = params.mask_size

        # `mask_size` binary polynomials
        key = IntPolynomialArray(rand_uniform_bool(thr, rng, (mask_size, polynomial_degree)))

        return cls(params, key)


class TLweSampleArray:

    def __init__(self, params: TLweParams, a, current_variances):
        self.a = a
        self.current_variances = current_variances
        self.shape = current_variances.shape
        self.params = params

    @classmethod
    def empty(cls, thr, params: TLweParams, shape):

        # array of length mask size + 1: mask + right term
        a = TorusPolynomialArray.empty(
            thr, params.polynomial_degree, shape + (params.mask_size + 1,))

        # avg variance of the sample
        current_variances = thr.to_device(numpy.zeros(shape, ErrorFloat))

        return cls(params, a, current_variances)


class TransformedTLweSampleArray:

    def __init__(self, params: TLweParams, a, current_variances):
        self.a = a
        self.current_variances = current_variances
        self.shape = current_variances.shape
        self.params = params

    @classmethod
    def empty(cls, thr, params: TLweParams, shape):

        # array of length mask size + 1: mask + right term
        a = TransformedPolynomialArray.empty(
            thr, params.transform_type, params.polynomial_degree, shape + (params.mask_size + 1,))

        # avg variance of the sample
        current_variances = thr.to_device(numpy.zeros(shape, ErrorFloat))

        return cls(params, a, current_variances)

    def dump(self, file_obj):
        pickle.dump(self.params, file_obj)
        self.a.dump(file_obj)
        pickle.dump(self.current_variances.get(), file_obj)

    @classmethod
    def load(cls, file_obj, thr):
        params = pickle.load(file_obj)
        a = TransformedPolynomialArray.load(file_obj, thr)
        current_variances = pickle.load(file_obj)
        return cls(params, a, thr.to_device(current_variances))

    def __eq__(self, other: 'TransformedTLweSampleArray'):
        return (
            self.__class__ == other.__class__
            and self.params == other.params
            and self.a == other.a
            and arrays_equal(self.current_variances, other.current_variances))


# result = (0,mu)
def tlwe_noiseless_trivial(thr, result: TLweSampleArray, mu: TorusPolynomialArray):
    comp = get_computation(thr, TLweNoiselessTrivial, result.params, result.shape)
    comp(result.a.coeffs, result.current_variances, mu.coeffs)


def tlwe_extract_lwe_samples(thr, result: LweSampleArray, x: TLweSampleArray):
    comp = get_computation(thr, TLweExtractLweSamples, x.params, x.shape)
    # Note: `current_variances` is not filled in original TFHE
    comp(result.a, result.b, x.a.coeffs)


# mult externe de X^ai-1 par bki
def tlwe_shift_polynomials(thr, result: TLweSampleArray, bk: TLweSampleArray, powers, powers_idx):
    shift_tp_minus_one_power_from_array(thr, result.a, powers, powers_idx, bk.a)


# result = result + sample
def tlwe_add_to(thr, result: TLweSampleArray, source: TLweSampleArray):
    result.a.coeffs += source.a.coeffs
    result.current_variances += source.current_variances


# result = sample
def tlwe_copy(thr, result: TLweSampleArray, source: TLweSampleArray):
    thr.copy(source.a.coeffs, dest=result.a.coeffs)
    thr.copy(source.current_variances, dest=result.current_variances)


# create an homogeneous tlwe sample
def tlwe_encrypt_zero(
        thr, rng, result: TLweSampleArray, noise: float, key: TLweKey,
        perf_params: PerformanceParametersForDevice):

    polynomial_degree = key.params.polynomial_degree
    mask_size = key.params.mask_size

    noises1 = rand_uniform_torus32(thr, rng, result.shape + (mask_size, polynomial_degree))
    noises2 = rand_gaussian_torus32(thr, rng, 0, noise, result.shape + (polynomial_degree,))

    comp = get_computation(thr, TLweEncryptZero, key.params, result.shape, noise, perf_params)
    comp(result.a.coeffs, result.current_variances, key.key.coeffs, noises1, noises2)


# Computes the inverse FFT of the coefficients of the TLWE sample
def tlwe_transform_samples(
        thr, result: TransformedTLweSampleArray, source: TLweSampleArray,
        perf_params: PerformanceParametersForDevice):

    comp = get_computation(
        thr, TLweTransformSamples, source.params, source.a.coeffs.shape, perf_params)
    comp(result.a.coeffs, source.a.coeffs)
    thr.copy_array(source.current_variances, dest=result.current_variances)



# Copyright (C) 2018 NuCypher
#
# This file is part of nufhe.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

import io
import pickle

import numpy

from .numeric_functions import phase_to_t32
from .lwe import LweParams, LweKey, LweSampleArray, lwe_encrypt, lwe_decrypt, LweKeyswitchKey
from .tgsw import TGswParams, TGswKey
from .tlwe import TLweParams
from .bootstrap import BootstrapKey
from .performance import PerformanceParameters, PerformanceParametersForDevice


class NuFHEParameters:
    """
    Parameters of the FHE scheme.

    :param transform_type: ``'NTT'`` or ``'FFT'``, specifying the transform to be used for
        internal purposes. ``'FFT'`` is generally faster, but may not be supported on
        some videocards (since it requires double precision floating point numbers).

    .. note::

        The default parameters correspond to about 128 bits of security.
    """

    def __init__(self, transform_type='NTT', tlwe_mask_size=1):

        assert transform_type in ('FFT', 'NTT')
        assert tlwe_mask_size >= 1

        tlwe_polynomial_degree = 1024
        lwe_size = 500

        bs_decomp_length = 2 # bootstrap decomposition length
        bs_log2_base = 10 # bootstrap log2(decomposition_base)

        ks_decomp_length = 8 # keyswitch decomposition length (the precision of the keyswitch)
        ks_log2_base = 2 # keyswitch log2(decomposition base)

        coeff = (2 / numpy.pi)**0.5
        ks_stdev = 1/2**15 * coeff # keyswitch minimal standard deviation
        bs_stdev = 9e-9 * coeff # bootstrap minimal standard deviation
        max_stdev = 1/2**4 / 4 * coeff # max standard deviation for a 1/4 msg space

        params_in = LweParams(lwe_size, ks_stdev, max_stdev)
        params_accum = TLweParams(
            tlwe_polynomial_degree, tlwe_mask_size, bs_stdev, max_stdev, transform_type)
        params_bs = TGswParams(params_accum, bs_decomp_length, bs_log2_base)

        self.ks_decomp_length = ks_decomp_length
        self.ks_log2_base = ks_log2_base
        self.in_out_params = params_in
        self.tgsw_params = params_bs

        self._transform_type = transform_type
        self._tlwe_mask_size = tlwe_mask_size

    def __hash__(self):
        return hash((
            self.__class__,
            self._transform_type,
            self._tlwe_mask_size,
            ))

    def __eq__(self, other: 'NuFHEParameters'):
        return (
            self.__class__ == other.__class__
            and self._transform_type == other._transform_type
            and self._tlwe_mask_size == other._tlwe_mask_size)


class NuFHESecretKey:
    """
    A secret key for the FHE scheme.

        .. py:attribute:: params

            A :py:class:`~nufhe.NuFHEParameters` object.
    """

    def __init__(self, params: NuFHEParameters, lwe_key: LweKey):
        """__init__()""" # hide the signature from Sphinx
        self.params = params
        self.lwe_key = lwe_key

    @classmethod
    def from_rng(cls, thr, params: NuFHEParameters, rng):
        """
        Generate a new secret key.

        :param thr: a ``reikna`` ``Thread`` object.
        :param params: FHE scheme parameters.
        :param rng: an RNG object, one of :ref:`random-number-generators`.
        """
        lwe_key = LweKey.from_rng(thr, params.in_out_params, rng)
        return cls(params, lwe_key)

    def dump(self, file_obj):
        """
        Serialize into the given ``file_obj``, a writeable file-like object.
        """
        pickle.dump(self.params, file_obj)
        self.lwe_key.dump(file_obj)

    def dumps(self):
        """
        Serialize into a bytestring.
        """
        file_obj = io.BytesIO()
        self.dump(file_obj)
        return file_obj.getvalue()

    @classmethod
    def load(cls, file_obj, thr):
        """
        Deserialize from the given ``file_obj``, a readable file-like object,
        using the ``reikna`` thread ``thr`` to store arrays.
        """
        params = pickle.load(file_obj)
        lwe_key = LweKey.load(file_obj, thr)
        return cls(params, lwe_key)

    @classmethod
    def loads(cls, s: bytes, thr):
        """
        Deserialize from the given bytestring
        using the ``reikna`` thread ``thr`` to store arrays.
        """
        file_obj = io.BytesIO(s)
        return cls.load(file_obj, thr)

    def __eq__(self, other: 'NuFHESecretKey'):
        return (
            self.__class__ == other.__class__
            and self.params == other.params
            and self.lwe_key == other.lwe_key)


class NuFHECloudKey:
    """
    A cloud key for the FHE scheme.

        .. py:attribute:: params

            A :py:class:`~nufhe.NuFHEParameters` object.
    """

    def __init__(
            self, params: NuFHEParameters,
            bootstrap_key: BootstrapKey, keyswitch_key: LweKeyswitchKey):
        """__init__()""" # hide the signature from Sphinx
        self.params = params
        self.bootstrap_key = bootstrap_key
        self.keyswitch_key = keyswitch_key

    @classmethod
    def from_rng(
            cls, thr, params: NuFHEParameters, rng, secret_key: NuFHESecretKey,
            perf_params: PerformanceParametersForDevice=None):
        """
        Generate a new cloud key based on the given secret key.

        :param thr: a ``reikna`` ``Thread`` object.
        :param params: FHE scheme parameters.
        :param rng: an RNG object, one of :ref:`random-number-generators`.
        :param secret_key: the secret key object.
        :param perf_params: an override for performance parameters.
        """

        if perf_params is None:
            perf_params = PerformanceParameters(params).for_device(thr.device_params)

        tgsw_key = TGswKey.from_rng(thr, params.tgsw_params, rng)
        bk = BootstrapKey.from_rng(thr, rng, secret_key.lwe_key, tgsw_key, perf_params)
        ks = LweKeyswitchKey.from_tgsw_key(
            thr, rng, params.ks_decomp_length, params.ks_log2_base,
            secret_key.lwe_key, tgsw_key)
        return cls(params, bk, ks)

    def dump(self, file_obj):
        """
        Serialize into the given ``file_obj``, a writeable file-like object.
        """
        pickle.dump(self.params, file_obj)
        self.bootstrap_key.dump(file_obj)
        self.keyswitch_key.dump(file_obj)

    def dumps(self):
        """
        Serialize into a bytestring.
        """
        file_obj = io.BytesIO()
        self.dump(file_obj)
        return file_obj.getvalue()

    @classmethod
    def load(cls, file_obj, thr):
        """
        Deserialize from the given ``file_obj``, a readable file-like object,
        using the ``reikna`` thread ``thr`` to store arrays.
        """
        params = pickle.load(file_obj)
        bootstrap_key = BootstrapKey.load(file_obj, thr)
        keyswitch_key = LweKeyswitchKey.load(file_obj, thr)
        return cls(params, bootstrap_key, keyswitch_key)

    @classmethod
    def loads(cls, s: bytes, thr):
        """
        Deserialize from the given bytestring
        using the ``reikna`` thread ``thr`` to store arrays.
        """
        file_obj = io.BytesIO(s)
        return cls.load(file_obj, thr)

    def __eq__(self, other: 'NuFHECloudKey'):
        return (
            self.__class__ == other.__class__
            and self.params == other.params
            and self.bootstrap_key == other.bootstrap_key
            and self.keyswitch_key == other.keyswitch_key)


def make_key_pair(thr, rng, **params):
    """
    Creates a pair of :py:class:`NuFHESecretKey` and :py:class:`NuFHECloudKey`
    corresponding to :py:class:`~nufhe.NuFHEParameters` created with keywords ``params``.
    """
    nufhe_params = NuFHEParameters(**params)
    secret_key = NuFHESecretKey.from_rng(thr, nufhe_params, rng)
    cloud_key = NuFHECloudKey.from_rng(thr, nufhe_params, rng, secret_key)
    return secret_key, cloud_key


_1s8 = phase_to_t32(1, 8)


@numpy.vectorize
def bool_to_t32(bit):
    return _1s8 if bit else -_1s8


@numpy.vectorize
def t32_to_bool(mu):
    return mu > 0


def encrypt(thr, rng, key: NuFHESecretKey, message):
    """
    Encrypts a message.

    :param rng: an RNG object, one of :ref:`random-number-generators`.
    :param key: the secret key.
    :param message: a ``numpy`` array of bit values to encrypt;
        if the ``dtype`` is not ``numpy.bool``, it will be converted to ``numpy.bool``.
    :returns: an :py:class:`LweSampleArray` object with the same `shape` as the given array.
    """
    message = numpy.asarray(message)
    result = empty_ciphertext(thr, key.params, message.shape)
    mus = thr.to_device(bool_to_t32(message))
    noise = key.params.in_out_params.min_noise
    lwe_encrypt(thr, rng, result, mus, noise, key.lwe_key)
    return result


def decrypt(thr, key: NuFHESecretKey, ciphertext: LweSampleArray):
    """
    Decrypts a message.

    :param key: the secret key.
    :param ciphertext: an encrypted message.
    :returns: a ``numpy.ndarray`` object of the type ``numpy.bool``
        and the same `shape` as ``ciphertext``.
    """

    mus = lwe_decrypt(thr, ciphertext, key.lwe_key)
    return t32_to_bool(mus)


def empty_ciphertext(thr, params: NuFHEParameters, shape):
    """
    Creates an uninitialized :py:class:`LweSampleArray` with the shape ``shape``.
    """
    return LweSampleArray.empty(thr, params.in_out_params, shape)



# Copyright (C) 2018 NuCypher
#
# This file is part of nufhe.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

import pickle

import numpy

from .utils import arrays_equal
from .computation_cache import get_computation
from .numeric_functions import Torus32, Int32
from .polynomial_transform import get_transform
from .polynomials_gpu import ShiftTorusPolynomial


# This structure represents an integer polynomial modulo X^N+1
class IntPolynomialArray:

    def __init__(self, coeffs):
        assert coeffs.dtype == Int32
        self.coeffs = coeffs
        self.shape = coeffs.shape[:-1]
        self.polynomial_degree = coeffs.shape[-1]


# This structure represents an torus polynomial modulo X^N+1
class TorusPolynomialArray:

    def __init__(self, coeffs):
        assert coeffs.dtype == Torus32
        self.coeffs = coeffs
        self.shape = coeffs.shape[:-1]
        self.polynomial_degree = coeffs.shape[-1]

    @classmethod
    def empty(cls, thr, polynomial_degree: int, shape):
        return cls(thr.array(shape + (polynomial_degree,), Torus32))


# Torus polynomial in transformed space
class TransformedPolynomialArray:

    def __init__(self, transform_type, coeffs):
        transform = get_transform(transform_type)
        assert coeffs.dtype == transform.transformed_dtype()
        self.transform_type = transform_type
        self.coeffs = coeffs
        self.shape = coeffs.shape[:-1]
        self.polynomial_degree = coeffs.shape[-1]

    @classmethod
    def empty(cls, thr, transform_type, polynomial_degree: int, shape):
        transform = get_transform(transform_type)
        coeffs = thr.array(
            shape + (transform.transformed_length(polynomial_degree),),
            transform.transformed_dtype())
        return cls(transform_type, coeffs)

    def dump(self, file_obj):
        pickle.dump(self.transform_type, file_obj)
        pickle.dump(self.coeffs.get(), file_obj)

    @classmethod
    def load(cls, file_obj, thr):
        transform_type = pickle.load(file_obj)
        coeffs = pickle.load(file_obj)
        return cls(transform_type, thr.to_device(coeffs))

    def __eq__(self, other):
        return (
            self.__class__ == other.__class__
            and self.transform_type == other.transform_type
            and arrays_equal(self.coeffs, other.coeffs))


# result = X^(2N - pwr) * source
def shift_tp_inverted_power(
        thr, result: TorusPolynomialArray, powers, source: TorusPolynomialArray):
    comp = get_computation(
        thr, ShiftTorusPolynomial,
        result.polynomial_degree, result.shape, powers.shape, invert_powers=True)
    comp(result.coeffs, source.coeffs, powers, 0)


# result = (X^pwr - 1) * source
def shift_tp_minus_one_power_from_array(
        thr, result: TorusPolynomialArray, powers, power_idx: int, source: TorusPolynomialArray):
    comp = get_computation(
        thr, ShiftTorusPolynomial,
        result.polynomial_degree, result.shape, powers.shape, powers_view=True, minus_one=True)
    comp(result.coeffs, source.coeffs, powers, power_idx)



# Copyright (C) 2018 NuCypher
#
# This file is part of nufhe.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

import numpy

import reikna.helpers as helpers
from reikna.cluda import dtypes, Module, functions


TEMPLATE = helpers.template_for(__file__)


def fft_transform_ref(data, inverse=False, i32_conversion=False):
    if i32_conversion and not inverse:
        N = data.shape[-1]
    else:
        N = data.shape[-1] * 2

    batch_shape = data.shape[:-1]
    data = data.reshape(numpy.prod(batch_shape), data.shape[-1])

    coeffs = numpy.exp(-2j * numpy.pi * numpy.arange(N // 2) / N / 2)

    f64_to_i32 = lambda x: numpy.round(x).astype(numpy.int64).astype(numpy.int32)

    if inverse:
        final_shape = batch_shape + ((N,) if i32_conversion else (N // 2,))
        res = numpy.fft.ifft(data).conj() * coeffs
        if i32_conversion:
            res = numpy.concatenate([
                f64_to_i32(res.real),
                f64_to_i32(res.imag)], axis=1)
        return res.reshape(final_shape)
    else:
        if i32_conversion:
            data = (data[:,:N//2] - 1j * data[:,N//2:])
        return numpy.fft.fft(data * coeffs).reshape(batch_shape + (N // 2,))


def fft_transformed_add_ref(data1, data2):
    return data1 + data2


def fft_transformed_mul_ref(data1, data2):
    return data1 * data2


class FFT512:

    def __init__(self, module, use_constant_memory):
        self.module = module
        self.use_constant_memory = use_constant_memory

        self.transform_length = 512
        self.elem_dtype = numpy.dtype('complex128')
        self.elem_ctype = dtypes.ctype(self.elem_dtype)

        self.polynomial_length = 1024
        self.polynomial_dtype = numpy.int32
        self.polynomial_ctype = dtypes.ctype(self.polynomial_dtype)

        self.threads_per_transform = 64
        self.temp_dtype = numpy.dtype('float64')
        self.temp_ctype = dtypes.ctype(self.temp_dtype)
        self.temp_length = 576

        twd_fw = numpy.empty((8, 64), numpy.complex128)
        twd_inv = numpy.empty((8, 64), numpy.complex128)
        for i in range(8):
            for elem_id in range(64):
                twd_fw[i, elem_id] = numpy.exp(
                    -2j * numpy.pi / self.transform_length * i * elem_id)
                twd_inv[i, elem_id] = numpy.exp(
                    2j * numpy.pi / self.transform_length * i * elem_id)

        idxs = numpy.arange(self.polynomial_length // 2)
        coeffs = numpy.exp(-2j * numpy.pi * idxs / self.polynomial_length / 2)

        self.cdata_fw = numpy.concatenate([twd_fw.flatten(), coeffs])
        self.cdata_inv = numpy.concatenate([twd_inv.flatten(), coeffs / self.transform_length])
        self.cdata_fw_ctype = dtypes.ctype(self.cdata_fw.dtype)
        self.cdata_inv_ctype = dtypes.ctype(self.cdata_inv.dtype)

    def __process_modules__(self, process):
        return FFT512(process(self.module), self.use_constant_memory)


def fft512(use_constant_memory=False):
    module = Module(
        TEMPLATE.get_def('fft512'),
        render_kwds=dict(
            elem_ctype=dtypes.ctype(numpy.complex128),
            temp_ctype=dtypes.ctype(numpy.float64),
            cdata_ctype=dtypes.ctype(numpy.complex128),
            polar_unit=functions.polar_unit(numpy.float64),
            mul=functions.mul(numpy.complex128, numpy.complex128),
            use_constant_memory=use_constant_memory,
            ))
    return FFT512(module, use_constant_memory)


def fft512_requirements():
    return dict(
        threads_per_transform=64,
        transform_length=512,
        temp_length=576,
        elem_dtype_itemsize=numpy.dtype('complex128').itemsize,
        temp_dtype_itemsize=numpy.dtype('float64').itemsize,
        polynomial_length=1024)



import numpy


def arrays_equal(arr1, arr2):
    """
    Returns ``True`` if two integer arrays are equal.
    """

    # PyOpenCL arrays have an ``all()`` method, but ``PyCUDA`` ones don't,
    # and they're both hidden behind Reikna ``Array`` anyway.
    # Since this function is used for testing purposes only,
    # we just bring any GPU array to CPU before comparison.

    if not isinstance(arr1, numpy.ndarray):
        arr1 = arr1.get()

    if not isinstance(arr2, numpy.ndarray):
        arr2 = arr2.get()

    return (arr1 == arr2).all()



# Copyright (C) 2018 NuCypher
#
# This file is part of nufhe.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

import numpy

import reikna.helpers as helpers
from reikna.cluda import dtypes, Module


TEMPLATE = helpers.template_for(__file__)


class FiniteFieldElement:

    def __init__(self, module):
        self.module = module
        self.u64 = dtypes.ctype(numpy.uint64)
        self.u32 = dtypes.ctype(numpy.uint32)
        self.modulus = dtypes.c_constant(2**64 - 2**32 + 1, numpy.uint64)

    def __process_modules__(self, process):
        return FiniteFieldElement(process(self.module))


def get_ff_elem():
    module = Module(
        TEMPLATE.get_def('ff_elem_def'),
        render_kwds=dict(u64=dtypes.ctype(numpy.uint64)))
    return FiniteFieldElement(module)


class FiniteFieldAdd:

    def __init__(self, ff_elem, module):
        self.ff = ff_elem
        self.module = module

    def __process_modules__(self, process):
        return FiniteFieldAdd(process(self.ff), process(self.module))


def add(ff_elem=None, method="c"):
    if ff_elem is None:
        ff_elem = get_ff_elem()
    module = Module(
        TEMPLATE.get_def('add_def'),
        render_kwds=dict(method=method, ff=ff_elem, ff_elem=ff_elem.module))
    return FiniteFieldAdd(ff_elem, module)


class FiniteFieldSub:

    def __init__(self, ff_elem, module):
        self.ff = ff_elem
        self.module = module

    def __process_modules__(self, process):
        return FiniteFieldSub(process(self.ff), process(self.module))


def sub(ff_elem=None, method="c"):
    if ff_elem is None:
        ff_elem = get_ff_elem()
    module = Module(
        TEMPLATE.get_def('sub_def'),
        render_kwds=dict(method=method, ff=ff_elem, ff_elem=ff_elem.module))
    return FiniteFieldSub(ff_elem, module)


class FiniteFieldMod:

    def __init__(self, ff_elem, module):
        self.ff = ff_elem
        self.module = module

    def __process_modules__(self, process):
        return FiniteFieldMod(process(self.ff), process(self.module))


def mod(ff_elem=None, method="c"):
    if ff_elem is None:
        ff_elem = get_ff_elem()
    module = Module(
        TEMPLATE.get_def('mod_def'),
        render_kwds=dict(method=method, ff=ff_elem, ff_elem=ff_elem.module))
    return FiniteFieldMod(ff_elem, module)


class FiniteFieldMul:

    def __init__(self, ff_elem, module):
        self.ff = ff_elem
        self.module = module

    def __process_modules__(self, process):
        return FiniteFieldMul(process(self.ff), process(self.module))


def mul(ff_elem=None, method="c", nested_method="c"):
    if ff_elem is None:
        ff_elem = get_ff_elem()
    module = Module(
        TEMPLATE.get_def('mul_def'),
        render_kwds=dict(
            method=method,
            ff=ff_elem, ff_elem=ff_elem.module,
            add=add(ff_elem=ff_elem, method=nested_method).module, # used for "c" method
            sub=sub(ff_elem=ff_elem, method=nested_method).module,
            mod=mod(ff_elem=ff_elem, method=nested_method).module, # used for "c_from_asm" method
            ))
    return FiniteFieldMul(ff_elem, module)


class FiniteFieldMulPrepared:

    def __init__(self, ff_elem, module):
        self.ff = ff_elem
        self.module = module

    def __process_modules__(self, process):
        return FiniteFieldMulPrepared(process(self.ff), process(self.module))


def mul_prepared(ff_elem=None, method="c", nested_method="c"):
    if ff_elem is None:
        ff_elem = get_ff_elem()
    module = Module(
        TEMPLATE.get_def('mul_prepared_def'),
        render_kwds=dict(
            method=method,
            ff=ff_elem, ff_elem=ff_elem.module,
            sub=sub(ff_elem=ff_elem, method=nested_method).module,
            ))
    return FiniteFieldMulPrepared(ff_elem, module)


class FiniteFieldPrepareForMul:

    def __init__(self, ff_elem, module):
        self.ff = ff_elem
        self.module = module

    def __process_modules__(self, process):
        return FiniteFieldPrepareForMul(process(self.ff), process(self.module))


def prepare_for_mul(ff_elem=None):
    if ff_elem is None:
        ff_elem = get_ff_elem()
    module = Module(
        TEMPLATE.get_def('prepare_for_mul_def'),
        render_kwds=dict(
            ff=ff_elem, ff_elem=ff_elem.module, sub=sub(ff_elem=ff_elem).module,
            ))
    return FiniteFieldPrepareForMul(ff_elem, module)


def prepare_for_mul_cpu(x):
    """
    Converts a numpy array `x` of type `uint64` into Montgomery represenation
    for modulus `M = 2**64-2**32+1` and word size `R = 2**64`.
    That is, returns `x * R mod M`.
    """

    # This is the CPU analogue of prepare_for_mul()

    # Using the known form of modulus, we are rewriting multiplications and modulo operations
    # as shifts to make it faster.
    #
    # The idea is as follows: `let r = 2**32` (so `r**2 = R`).
    # Then `(a + b * r) * r**2 mod (r**2 - r + 1) = a * r - a - b mod (r**2 - r + 1)`.
    # `a * r - a` is always in range `[0, m)`, so we just need to check if it is
    # smaller than `b` and, in that case, add `m` (which is equivalent to subtracting `r - 1`
    # with modulo `r**2` applied automatically since we're working with `uint64`).

    x_lo = x & numpy.uint64(0xffffffff)
    x_hi = x >> 32
    y = (x_lo << 32) - x_lo
    adjust = (y < x_hi).astype(numpy.uint64)
    adjust = (adjust << 32) - adjust
    return y - x_hi - adjust


class FiniteFieldPow:

    def __init__(self, ff_elem, module, exp_dtype):
        self.ff = ff_elem
        self.module = module
        self.exp_dtype = exp_dtype

    def __process_modules__(self, process):
        return FiniteFieldPow(process(self.ff), process(self.module), self.exp_dtype)


def pow(exp_dtype, ff_elem=None):
    if ff_elem is None:
        ff_elem = get_ff_elem()
    module = Module(
        TEMPLATE.get_def('pow_def'),
        render_kwds=dict(
            ff=ff_elem, ff_elem=ff_elem.module, mul=mul(ff_elem).module, exp_dtype=exp_dtype))
    return FiniteFieldPow(ff_elem, module, exp_dtype)


class FiniteFieldInvPow2:

    def __init__(self, ff_elem, module, exp_dtype):
        self.ff = ff_elem
        self.module = module
        self.exp_dtype = exp_dtype

    def __process_modules__(self, process):
        return FiniteFieldInvPow2(process(self.ff), process(self.module), self.exp_dtype)


def inv_pow2(exp_dtype, ff_elem=None):
    # this can be relaxed when we witch to regular C
    assert numpy.dtype(exp_dtype) == numpy.dtype('uint32')

    if ff_elem is None:
        ff_elem = get_ff_elem()
    module = Module(
        TEMPLATE.get_def('inv_pow2_def'),
        render_kwds=dict(ff=ff_elem, ff_elem=ff_elem.module, exp_dtype=exp_dtype))
    return FiniteFieldPow(ff_elem, module, exp_dtype)


class FiniteFieldLsh:

    def __init__(self, ff_elem, module, exp_dtype):
        self.ff = ff_elem
        self.module = module
        self.exp_dtype = exp_dtype

    def __process_modules__(self, process):
        return FiniteFieldLsh(process(self.ff), process(self.module), self.exp_dtype)


def lsh(exp_range, exp_dtype, ff_elem=None, method="c", nested_method="c"):
    # this can be relaxed when we witch to regular C
    assert numpy.dtype(exp_dtype) == numpy.dtype('uint32')

    assert exp_range in (32, 64, 96, 128, 160, 192)

    if ff_elem is None:
        ff_elem = get_ff_elem()
    module = Module(
        TEMPLATE.get_def('lsh_def'),
        render_kwds=dict(
            method=method,
            sub=sub(ff_elem=ff_elem, method=nested_method).module, # used in "c" method
            add=add(ff_elem=ff_elem, method=nested_method).module, # used in "c" method
            mod=mod(ff_elem=ff_elem, method=nested_method).module, # used in "c_from_asm" method
            exp_range=exp_range, ff=ff_elem, ff_elem=ff_elem.module,
            exp_dtype=exp_dtype))
    return FiniteFieldLsh(ff_elem, module, exp_dtype)



# Copyright (C) 2018 NuCypher
#
# This file is part of nufhe.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

"""
At the moment the random numer generating functions simply use a CPU RNG
and transfer the result to GPU. The reasons are:

- random number generation is fast compared to the rest of the algorithms
- a GPU RNG requires some non-trivial logistics with preserving/creating the random state
- it makes it is easier to test the rest of the code during the transition from CPU to GPU

When it is necessary, the functions can be made to execute on GPU without changing the API.
"""

import numpy
import random

from os import urandom
from reikna.helpers import min_blocks

from .numeric_functions import double_to_t32, Torus32, Int32


# Used to generate uniform floats.
Float = numpy.dtype('float64')
MantissaInt = numpy.dtype('uint64') # an integer type that fits all possible mantissa values

BPF = numpy.finfo(Float).nmant + 1 # mantissa size + the implicit constant bit
RECIP_BPF = 2**(-BPF)


class DeterministicRNG:
    """
    A fast, but not cryptographically secure RNG.
    Useful for testing, since it allows seeding the initial state.
    """

    def __init__(self, seed=None):
        self.rng = numpy.random.RandomState(seed)

    def uniform_bool(self, shape):
        return self.rng.randint(0, 2, size=shape, dtype=Int32)

    def uniform_torus32(self, shape):
        return self.rng.randint(-2**31, 2**31, size=shape, dtype=Torus32)

    def gauss(self, shape, std_dev):
        return self.rng.normal(size=shape, scale=std_dev)


class SecureRNG:
    """
    A cryptographically secure RNG provided by the OS.

    .. note::

        This RNG can be very slow, leading to cloud key creation times of the order of minutes.
        Encryption is not affected too much (the required amount of random numbers is much lower).
    """

    def __init__(self):
        self.rng = random.SystemRandom()

    def uniform_bool(self, shape):
        length = numpy.prod(shape)
        nbytes = min_blocks(length, 8)
        random_bytes = numpy.frombuffer(urandom(nbytes), numpy.uint8)
        random_bits = numpy.unpackbits(random_bytes)[:length]
        return random_bits.reshape(shape).astype(Int32)

    def uniform_torus32(self, shape):
        length = numpy.prod(shape)
        nbytes = length * numpy.dtype(Int32).itemsize
        return numpy.frombuffer(urandom(nbytes), Int32).reshape(shape)

    def _uniform_float(self, length):
        """
        Returns an array of uniformly distributed floats in the interval (0, 1)
        (open at both ends!).
        """

        # The number of possible values in the interval [0, 1) is
        # 2**(mantissa size + the implicit constant bit)
        # (and they are uniformly distributed).

        nbytes = length * MantissaInt.itemsize
        mantissa_bits = numpy.frombuffer(urandom(nbytes), MantissaInt)

        # We want an open interval, so we want to exclude 0 without using rejection sampling.
        # To do that, we drop an additional bit from our generated integer,
        # and shift the range from [0, 2^(bpf-1)-1] to [2^(-bpf), 1 - 2^(-bpf)]
        mantissa_bits = mantissa_bits >> (MantissaInt.itemsize * 8 - (BPF - 1))
        mantissa_bits = mantissa_bits * 2 + 1

        return mantissa_bits * RECIP_BPF

    def gauss(self, shape, std_dev):
        orig_length = numpy.prod(shape)

        # Make the length even since Box-Muller transform generates pairs of randoms
        length = orig_length + orig_length % 2

        # Box-Muller transform

        u1 = self._uniform_float(length // 2)
        u2 = self._uniform_float(length // 2)

        r = (-2 * numpy.log(u1))**0.5
        theta = 2 * numpy.pi * u2

        z0 = r * numpy.cos(theta)
        z1 = r * numpy.sin(theta)

        result = numpy.concatenate([z0, z1])[:orig_length]

        return result.reshape(shape) * std_dev


# Gaussian sample centered in message, with standard deviation sigma
def _rand_gaussian_torus32(rng, message: Torus32, sigma: float, shape, centered=False):
    # Attention: all the implementation will use the stdev instead of the gaussian fourier param
    rfloats = rng.gauss(shape, sigma)
    if centered:
        rfloats -= rfloats.mean()
    return Torus32(message) + double_to_t32(rfloats)


def rand_uniform_bool(thr, rng, shape):
    return thr.to_device(rng.uniform_bool(shape))


def rand_uniform_torus32(thr, rng, shape):
    return thr.to_device(rng.uniform_torus32(shape))


def rand_gaussian_torus32(thr, rng, message: Torus32, sigma: float, shape, centered=False):
    return thr.to_device(_rand_gaussian_torus32(rng, message, sigma, shape, centered=centered))



# Copyright (C) 2018 NuCypher
#
# This file is part of nufhe.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

import numpy

from reikna.cluda import functions, Module
from reikna.core import Transformation, Parameter, Annotation, Type

from .transform import fft512, fft512_requirements, Transform
from .transform.fft import fft_transform_ref
from .performance import PerformanceParametersForDevice


def transformed_dtype():
    return numpy.dtype('complex128')


def transformed_internal_dtype():
    return numpy.dtype('complex128')


elem =  Module.create(
    lambda prefix: """
    typedef double2 ${prefix};

    #define ${prefix}pack(x) (x)
    #define ${prefix}unpack(x) (x)
    #define ${prefix}zero (COMPLEX_CTR(double2)(0, 0))
    """,
    render_kwds=dict())


def transformed_internal_ctype():
    return elem


def transformed_length(N):
    return N // 2


def forward_transform_ref(data):
    return fft_transform_ref(data, i32_conversion=True)


def inverse_transform_ref(data):
    return fft_transform_ref(data, i32_conversion=True, inverse=True)


def transformed_space_add_ref(data1, data2):
    return data1 + data2


def transformed_space_mul_ref(data1, data2):
    return data1 * data2


def transformed_space_mul_prepared_ref(data1, data2):
    return transformed_space_mul_ref(data1, data2)


def transformed_add(perf_params):
    return functions.add(transformed_dtype(), transformed_dtype())


def transformed_mul(perf_params):
    return functions.mul(transformed_dtype(), transformed_dtype())


def transformed_mul_prepared(perf_params):
    return transformed_mul(perf_params)


def transform_module_requirements():
    return fft512_requirements()


def get_prepare_for_mul_trf(shape):
    # Preparation for FFT is just an identity
    dtype = transformed_dtype()
    return Transformation([
        Parameter('output', Annotation(Type(dtype, shape), 'o')),
        Parameter('input', Annotation(Type(dtype, shape), 'i'))],
        """
        ${output.store_same}(${input.load_same});
        """,
        connectors=['input', 'output'])


def transform_module(perf_params: PerformanceParametersForDevice, multi_iter=False):
    use_constant_memory = (
        perf_params.use_constant_memory_multi_iter if multi_iter
        else perf_params.use_constant_memory_single_iter)
    return fft512(use_constant_memory=use_constant_memory)


def ForwardTransform(batch_shape, N, perf_params: PerformanceParametersForDevice):
    assert N == 1024
    return Transform(
        transform_module(perf_params), batch_shape,
        transforms_per_block=perf_params.transforms_per_block, i32_conversion=True)


def InverseTransform(batch_shape, N, perf_params: PerformanceParametersForDevice):
    assert N == 1024
    return Transform(
        transform_module(perf_params), batch_shape,
        transforms_per_block=perf_params.transforms_per_block, i32_conversion=True, inverse=True)



# Copyright (C) 2018 NuCypher
#
# This file is part of nufhe.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

import numpy

from nufhe.api_low_level import NuFHEParameters
from nufhe.numeric_functions import Torus32, Int32, ErrorFloat
from nufhe.tlwe import TLweParams
from nufhe.tlwe_gpu import (
    TLweNoiselessTrivial,
    TLweExtractLweSamples,
    TLweEncryptZero,
    )
from nufhe.tlwe_cpu import (
    TLweNoiselessTrivialReference,
    TLweExtractLweSamplesReference,
    TLweEncryptZeroReference,
    )
from nufhe import PerformanceParameters

from utils import get_test_array, errors_allclose


def test_tlwe_noiseless_trivial(thread):

    shape = (2, 5)
    params = NuFHEParameters().tgsw_params.tlwe_params

    mask_size = params.mask_size
    polynomial_degree = params.polynomial_degree

    mu = get_test_array(shape + (polynomial_degree,), Torus32)
    a = numpy.empty(shape + (mask_size + 1, polynomial_degree), Torus32)
    cv = numpy.empty(shape, ErrorFloat)

    a_dev = thread.empty_like(a)
    cv_dev = thread.empty_like(cv)
    mu_dev = thread.to_device(mu)

    test = TLweNoiselessTrivial(params, shape).compile(thread)
    ref = TLweNoiselessTrivialReference(params, shape)

    test(a_dev, cv_dev, mu_dev)
    a_test = a_dev.get()
    cv_test = cv_dev.get()

    ref(a, cv, mu)

    assert (a_test == a).all()
    assert errors_allclose(cv_test, cv)


def test_tlwe_extract_lwe_samples(thread):

    shape = (2, 5)
    params = NuFHEParameters().tgsw_params.tlwe_params

    mask_size = params.mask_size
    polynomial_degree = params.polynomial_degree

    tlwe_a = get_test_array(shape + (mask_size + 1, polynomial_degree), Torus32)

    a = numpy.empty(shape + (params.extracted_lweparams.size,), Torus32)
    b = numpy.empty(shape, Torus32)

    tlwe_a_dev = thread.to_device(tlwe_a)
    a_dev = thread.empty_like(a)
    b_dev = thread.empty_like(b)

    test = TLweExtractLweSamples(params, shape).compile(thread)
    ref = TLweExtractLweSamplesReference(params, shape)

    test(a_dev, b_dev, tlwe_a_dev)
    a_test = a_dev.get()
    b_test = b_dev.get()

    ref(a, b, tlwe_a)

    assert (a_test == a).all()
    assert (b_test == b).all()


def test_tlwe_encrypt_zero(thread):

    nufhe_params = NuFHEParameters()
    perf_params = PerformanceParameters(nufhe_params).for_device(thread.device_params)
    params = nufhe_params.tgsw_params.tlwe_params

    mask_size = params.mask_size
    polynomial_degree = params.polynomial_degree
    noise = params.min_noise

    shape = (3, 4, 5)

    result_a = numpy.empty(shape + (mask_size + 1, polynomial_degree), Torus32)
    result_cv = numpy.empty(shape, ErrorFloat)
    noises1 = get_test_array(shape + (mask_size, polynomial_degree), Torus32)
    noises2 = get_test_array(shape + (polynomial_degree,), Torus32)
    key = get_test_array((mask_size, polynomial_degree), Int32, (0, 2))

    test = TLweEncryptZero(params, shape, noise, perf_params).compile(thread)
    ref = TLweEncryptZeroReference(params, shape, noise, perf_params)

    result_a_dev = thread.empty_like(result_a)
    result_cv_dev = thread.empty_like(result_cv)
    noises1_dev = thread.to_device(noises1)
    noises2_dev = thread.to_device(noises2)
    key_dev = thread.to_device(key)

    test(result_a_dev, result_cv_dev, key_dev, noises1_dev, noises2_dev)
    ref(result_a, result_cv, key, noises1, noises2)

    result_a_test = result_a_dev.get()
    result_cv_test = result_cv_dev.get()

    assert (result_a_test == result_a).all()
    assert errors_allclose(result_cv_test, result_cv)



# Copyright (C) 2018 NuCypher
#
# This file is part of nufhe.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

import numpy

from .numeric_functions_gpu import Torus32, Int32


def Torus32ToPhaseReference(shape, mspace_size: int):

    interv = numpy.uint32(2**32 // mspace_size)
    half_interv = numpy.uint32(interv // 2)

    def _kernel(result, phase):

        nonlocal interv

        assert phase.dtype == Torus32
        assert result.dtype == Int32

        numpy.copyto(result, ((phase.astype(numpy.uint32) + half_interv) // interv).astype(Int32))

    return _kernel



# Copyright (C) 2018 NuCypher
#
# This file is part of nufhe.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

import numpy

from reikna.core import Transformation, Parameter, Annotation, Type

from .transform import ntt1024, ntt1024_requirements, Transform
from .transform.arithmetic import add, mul, get_ff_elem, prepare_for_mul, mul_prepared
from .transform.ntt import ntt_transform_ref
from .transform import ntt_cpu
from .performance import PerformanceParametersForDevice


def transformed_dtype():
    return numpy.dtype('uint64')


def transformed_internal_dtype():
    return numpy.dtype([("val", numpy.uint64)])


def transformed_internal_ctype():
    return ff_elem.module


def transformed_length(N):
    return N


def forward_transform_ref(data):
    return ntt_transform_ref(data, i32_conversion=True)


def inverse_transform_ref(data):
    return ntt_transform_ref(data, i32_conversion=True, inverse=True)


def transformed_space_add_ref(data1, data2):
    data1 = ntt_cpu.gnum(data1)
    data2 = ntt_cpu.gnum(data2)
    return ntt_cpu.gnum_to_u64(data1 + data2)


def transformed_space_mul_ref(data1, data2):
    data1 = ntt_cpu.gnum(data1)
    data2 = ntt_cpu.gnum(data2)
    return ntt_cpu.gnum_to_u64(data1 * data2)


def transformed_space_mul_prepared_ref(data1, data2):
    coeff = ntt_cpu.gnum(0xfffffffe00000001) # Inverse of 2**64 modulo (2**64-2**32+1)
    data1 = ntt_cpu.gnum(data1)
    data2 = ntt_cpu.gnum(data2)
    return ntt_cpu.gnum_to_u64(data1 * data2 * coeff)


ff_elem = get_ff_elem()


def transformed_add(perf_params: PerformanceParametersForDevice):
    return add(ff_elem=ff_elem, method=perf_params.ntt_base_method).module


def transformed_mul(perf_params: PerformanceParametersForDevice):
    return mul(ff_elem=ff_elem, method=perf_params.ntt_mul_method).module


def transformed_mul_prepared(perf_params: PerformanceParametersForDevice):
    return mul_prepared(ff_elem=ff_elem, method=perf_params.ntt_mul_method).module


def transform_module_requirements():
    return ntt1024_requirements()


def get_prepare_for_mul_trf(shape):
    dtype = transformed_dtype()
    return Transformation([
        Parameter('output', Annotation(Type(dtype, shape), 'o')),
        Parameter('input', Annotation(Type(dtype, shape), 'i'))],
        """
        ${dtypes.ctype(dtype)} x = ${input.load_same};
        ${ff_ctype} x_ff = { x };
        ${output.store_same}(${prepare_for_mul}(x_ff).val);
        """,
        connectors=['input', 'output'],
        render_kwds=dict(
            prepare_for_mul=prepare_for_mul(ff_elem=ff_elem).module,
            dtype=dtype,
            ff_ctype=transformed_internal_ctype()))


def transform_module(perf_params: PerformanceParametersForDevice, multi_iter=False):
    use_constant_memory = (
        perf_params.use_constant_memory_multi_iter if multi_iter
        else perf_params.use_constant_memory_single_iter)
    return ntt1024(
        ff_elem=ff_elem,
        base_method=perf_params.ntt_base_method,
        mul_method=perf_params.ntt_mul_method,
        lsh_method=perf_params.ntt_lsh_method,
        use_constant_memory=use_constant_memory)


def ForwardTransform(batch_shape, N, perf_params: PerformanceParametersForDevice):
    assert N == 1024
    return Transform(
        transform_module(perf_params), batch_shape,
        transforms_per_block=perf_params.transforms_per_block, i32_conversion=True)


def InverseTransform(batch_shape, N, perf_params: PerformanceParametersForDevice):
    assert N == 1024
    return Transform(
        transform_module(perf_params), batch_shape,
        transforms_per_block=perf_params.transforms_per_block, i32_conversion=True, inverse=True)



# Copyright (C) 2018 NuCypher
#
# This file is part of nufhe.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

import time
import numpy
import pytest

import reikna.cluda as cluda
import reikna.cluda.dtypes as dtypes
import reikna.cluda.functions as functions
from reikna.helpers import product

import nufhe.transform as ntt
import nufhe.transform.ntt_cpu as ntt_cpu

from utils import get_test_array, tp_dtype


@pytest.fixture(params=["c", "cuda_asm", "c_from_asm"])
def method(request):
    return request.param


def get_func_kernel(thread, func_module, output_type, input_types, repetitions=1):
    src = """
    <%
        argnames = ["a" + str(i + 1) for i in range(len(input_types))]
        ctype = lambda x: str(func.ff.module) if x == 'ff_number' else dtypes.ctype(x)
    %>
    KERNEL void test(
        GLOBAL_MEM ${ctype(output_type)} *dest
        %for arg, tp in zip(argnames, input_types):
        , GLOBAL_MEM ${ctype(tp)} *${arg}
        %endfor
        )
    {
        const SIZE_T i = get_global_id(0);
        %for arg, tp in zip(argnames, input_types):
        ${ctype(tp)} ${arg}_load = ${arg}[i];
        %endfor

        // To stop the compiler from optimizing away the code,
        // we need to use the return value somehow.
        // We're using it as the first argument to the next invocation.
        // Assuming here that the first argument of the tested function is either
        // a finite field element, or something convertable to it (e.g. uint64).
        ${ctype(output_type)} res =
        %if ctype(input_types[0]) == ctype(output_type):
            ${argnames[0]}_load;
        %else:
            { ${argnames[0]}_load };
        %endif

        for (int i = 0; i < ${repetitions}; i++)
        {
            res = ${func.module}(
                %if ctype(input_types[0]) == ctype(output_type):
                res
                %else:
                res.val
                %endif
                %if len(argnames) > 1:
                , ${", ".join([arg + "_load" for arg in argnames[1:]])}
                %endif
                );
        }

        dest[i] = res;
    }
    """

    program = thread.compile(
        src,
        render_kwds=dict(
            input_types=input_types, output_type=output_type, func=func_module,
            repetitions=repetitions))

    return program.test


def check_func(
        thread, func_module, reference_func, output_type, input_types,
        ranges=None, test_values=None):

    N = 1024

    test = get_func_kernel(thread, func_module, output_type, input_types)

    arrays = [
        get_test_array(N, tp, val_range=ranges[i] if ranges is not None else None)
        for i, tp in enumerate(input_types)]

    if test_values is not None:
        for i, tvs in enumerate(test_values):
            if tvs is not None:
                for j, tv in enumerate(tvs):
                    arrays[j][i] = tv

    arrays_dev = [thread.to_device(arr) for arr in arrays]
    dest_dev = thread.array(N, tp_dtype(output_type))

    test(dest_dev, *arrays_dev, global_size=N)

    assert (dest_dev.get() == reference_func(*arrays)).all()


def ref_add(data1, data2):
    data1 = ntt_cpu.gnum(data1)
    data2 = ntt_cpu.gnum(data2)
    return ntt_cpu.gnum_to_u64(data1 + data2)


def test_add(thread, method):
    if method == "cuda_asm" and thread.api.get_id() != cluda.cuda_id():
        pytest.skip()
    check_func(thread, ntt.add(method=method), ref_add, 'ff_number', ['ff_number', 'ff_number'])


def ref_sub(data1, data2):
    data1 = ntt_cpu.gnum(data1)
    data2 = ntt_cpu.gnum(data2)
    return ntt_cpu.gnum_to_u64(data1 - data2)


def test_sub(thread, method):
    if method == "cuda_asm" and thread.api.get_id() != cluda.cuda_id():
        pytest.skip()
    check_func(thread, ntt.sub(method=method), ref_sub, 'ff_number', ['ff_number', 'ff_number'])


def ref_mod(data):
    return data % ntt_cpu.GaloisNumber.modulus


def test_mod(thread, method):
    if method == "cuda_asm" and thread.api.get_id() != cluda.cuda_id():
        pytest.skip()
    check_func(
        thread, ntt.mod(method=method), ref_mod, 'ff_number', [numpy.uint64],
        test_values=[
            (ntt_cpu.GaloisNumber.modulus - 1,),
            (ntt_cpu.GaloisNumber.modulus,),
            (ntt_cpu.GaloisNumber.modulus + 1,)])


def ref_mul(data1, data2):
    data1 = ntt_cpu.gnum(data1)
    data2 = ntt_cpu.gnum(data2)
    return ntt_cpu.gnum_to_u64(data1 * data2)


def test_mul(thread, method):
    if method == "cuda_asm" and thread.api.get_id() != cluda.cuda_id():
        pytest.skip()
    check_func(
        thread, ntt.mul(method=method), ref_mul, 'ff_number', ['ff_number', 'ff_number'],
        test_values=[
            (ntt_cpu.GaloisNumber.modulus - 1, 2**33) # regression test for an error in method=c
            ]
        )


def ref_mul_prepared(data1, data2):
    coeff = ntt_cpu.gnum(0xfffffffe00000001) # Inverse of 2**64 modulo (2**64-2**32+1)
    data1 = ntt_cpu.gnum(data1)
    data2 = ntt_cpu.gnum(data2)
    return ntt_cpu.gnum_to_u64(data1 * data2 * coeff)


def test_mul_prepared(thread, method):
    if method == "cuda_asm" and thread.api.get_id() != cluda.cuda_id():
        pytest.skip()
    check_func(
        thread, ntt.mul_prepared(method=method), ref_mul_prepared,
        'ff_number', ['ff_number', 'ff_number'])


def ref_prepare_for_mul(data1):
    coeff = ntt_cpu.gnum(0xffffffff) # 2**64 modulo (2**64-2**32+1)
    data1 = ntt_cpu.gnum(data1)
    return ntt_cpu.gnum_to_u64(data1 * coeff)


def test_prepare_for_mul(thread):
    check_func(
        thread, ntt.prepare_for_mul(), ref_prepare_for_mul,
        'ff_number', ['ff_number'])


def test_prepare_for_mul_cpu():
    array = get_test_array(1024, 'ff_number')
    res = ntt.prepare_for_mul_cpu(array)
    ref = ref_prepare_for_mul(array)
    assert (res == ref).all()


def ref_pow(data1, data2):
    data1 = ntt_cpu.gnum(data1)
    return ntt_cpu.gnum_to_u64(data1**data2)


def test_pow(thread):
    exp_dtype = numpy.uint32
    check_func(thread, ntt.pow(exp_dtype), ref_pow, 'ff_number', ['ff_number', exp_dtype])


def ref_inv_pow2(data1):
    return ntt_cpu.gnum_to_u64(ntt_cpu.gnum(1) / ntt_cpu.gnum(2)**data1)


def test_inv_pow2(thread):
    exp_dtype = numpy.uint32
    check_func(
        thread, ntt.inv_pow2(exp_dtype), ref_inv_pow2, 'ff_number', [exp_dtype],
        ranges=[(1, 33)], test_values=[(1,), (32,)])


def ref_lsh(data1, data2):
    data1 = ntt_cpu.gnum(data1)
    data2 = ntt_cpu.gnum(2)**data2
    return ntt_cpu.gnum_to_u64(data1 * data2)


@pytest.mark.parametrize("exp_range", [32, 64, 96, 128, 160, 192])
def test_lsh(thread, exp_range, method):
    if method == "cuda_asm" and thread.api.get_id() != cluda.cuda_id():
        pytest.skip()
    exp_dtype = numpy.uint32
    check_func(
        thread, ntt.lsh(exp_range, exp_dtype, method=method),
        ref_lsh, 'ff_number', ['ff_number', exp_dtype],
        ranges=[None, (exp_range - 32, exp_range)],
        test_values=[
            (11509900421665959066, exp_range - 1)
        ])


def check_func_performance(
        tag, thread, func_module, reference_func, output_type, input_types, ranges=None,
        heavy_performance_load=False):

    N = 1024 * (1024 if heavy_performance_load else 32)
    repetitions = 100000

    test = get_func_kernel(thread, func_module, output_type, input_types)
    perf_test = get_func_kernel(
        thread, func_module, output_type, input_types, repetitions=repetitions)

    arrays = [
        get_test_array(N, tp, val_range=ranges[i] if ranges is not None else None)
        for i, tp in enumerate(input_types)]

    arrays_dev = [thread.to_device(arr) for arr in arrays]
    dest_dev = thread.array(N, tp_dtype(output_type))

    # Sanity check
    test(dest_dev, *arrays_dev, global_size=N)
    assert (dest_dev.get() == reference_func(*arrays)).all()

    # Performance check
    times = []

    for j in range(10):
        thread.synchronize()
        t1 = time.time()
        perf_test(dest_dev, *arrays_dev, global_size=N)
        thread.synchronize()
        t2 = time.time()
        times.append(t2 - t1)

    times = numpy.array(times)

    times /= repetitions
    times /= N
    times *= 1e12

    print()
    print(
        "{backend}: {tag} --- min: {min:.4f}, mean: {mean:.4f}, std: {std:.4f}".format(
            tag=tag,
            min=times.min(), mean=times.mean(), std=times.std(),
            backend='cuda' if thread.api.get_id() == cluda.cuda_id() else 'ocl '))


@pytest.mark.perf
def test_add_perf(thread, method, heavy_performance_load):
    if method == "cuda_asm" and thread.api.get_id() != cluda.cuda_id():
        pytest.skip()
    if method == "c_from_asm":
        pytest.skip()

    check_func_performance(
        "add(), " + method,
        thread, ntt.add(method=method), ref_add, 'ff_number', ['ff_number', 'ff_number'],
        heavy_performance_load=heavy_performance_load)


@pytest.mark.perf
def test_sub_perf(thread, method, heavy_performance_load):
    if method == "cuda_asm" and thread.api.get_id() != cluda.cuda_id():
        pytest.skip()
    if method == "c_from_asm":
        pytest.skip()

    check_func_performance(
        "sub(), " + method,
        thread, ntt.sub(method=method), ref_sub, 'ff_number', ['ff_number', 'ff_number'],
        heavy_performance_load=heavy_performance_load)



# Copyright (C) 2018 NuCypher
#
# This file is part of nufhe.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

import time

import pytest
import numpy

from reikna.cluda import cuda_id

from nufhe import *
from nufhe.operators_integer import uint_min, bitarray_to_uintarray, uintarray_to_bitarray
from nufhe.blind_rotate import single_kernel_bootstrap_supported
from nufhe.polynomial_transform import max_supported_transforms_per_block, transform_supported


@pytest.fixture(scope='module', params=[False, True], ids=['bs_loop', 'bs_kernel'])
def single_kernel_bootstrap(request):
    return request.param


def get_plaintexts(rng, num, shape=(32,)):
    return [rng.uniform_bool(shape).astype(numpy.bool) for i in range(num)]


def check_gate(
        thread, key_pair, num_arguments, nufhe_func, reference_func,
        shape=32, performance_test=False, perf_params=None):

    if not isinstance(shape, tuple):
        shape = (shape,)

    secret_key, cloud_key = key_pair

    if perf_params is None:
        perf_params = PerformanceParameters(secret_key.params).for_device(thread.device_params)

    rng = DeterministicRNG()

    plaintexts = get_plaintexts(rng, num_arguments, shape=shape)
    ciphertexts = [encrypt(thread, rng, secret_key, plaintext) for plaintext in plaintexts]

    reference = reference_func(*plaintexts)

    params = cloud_key.params
    answer = empty_ciphertext(thread, params, shape)

    if performance_test:

        # warm-up
        nufhe_func(thread, cloud_key, answer, *ciphertexts, perf_params)
        thread.synchronize()

        # test
        times = []
        for i in range(10):
            t_start = time.time()
            nufhe_func(thread, cloud_key, answer, *ciphertexts, perf_params)
            thread.synchronize()
            times.append(time.time() - t_start)
        times = numpy.array(times)

    else:
        nufhe_func(thread, cloud_key, answer, *ciphertexts, perf_params)
        times = None

    answer_bits = decrypt(thread, secret_key, answer)

    assert (answer_bits == reference).all()

    return times


def test_transform_type(thread, transform_type):
    if not transform_supported(thread.device_params, transform_type):
        pytest.skip()
    rng = DeterministicRNG()
    key_pair = make_key_pair(thread, rng, transform_type=transform_type)
    check_gate(thread, key_pair, 2, gate_nand, nand_ref)


@pytest.mark.parametrize('tlwe_mask_size', [1, 2], ids=['mask_size=1', 'mask_size=2'])
def test_tlwe_mask_size(thread, tlwe_mask_size):
    rng = DeterministicRNG()
    secret_key, cloud_key = make_key_pair(thread, rng, tlwe_mask_size=tlwe_mask_size)
    check_gate(thread, (secret_key, cloud_key), 2, gate_nand, nand_ref)


def test_single_kernel_bs_with_ks(thread, key_pair, single_kernel_bootstrap):
    # Test a gate that employs a bootstrap with keyswitch
    secret_key, cloud_key = key_pair

    if (single_kernel_bootstrap
            and not single_kernel_bootstrap_supported(secret_key.params, thread.device_params)):
        pytest.skip()

    perf_params = PerformanceParameters(
        secret_key.params, single_kernel_bootstrap=single_kernel_bootstrap)
    perf_params = perf_params.for_device(thread.device_params)
    check_gate(thread, key_pair, 2, gate_nand, nand_ref, perf_params=perf_params)


def test_single_kernel_bs(thread, key_pair, single_kernel_bootstrap):
    # Test a gate that employs separate calls to bootstrap and keyswitch
    secret_key, cloud_key = key_pair

    if (single_kernel_bootstrap
            and not single_kernel_bootstrap_supported(secret_key.params, thread.device_params)):
        pytest.skip()

    perf_params = PerformanceParameters(
        secret_key.params, single_kernel_bootstrap=single_kernel_bootstrap)
    perf_params = perf_params.for_device(thread.device_params)
    check_gate(thread, key_pair, 3, gate_mux, mux_ref, perf_params=perf_params)


def nand_ref(a, b):
    return ~(a * b)

def or_ref(a, b):
    return a + b

def and_ref(a, b):
    return a * b

def xor_ref(a, b):
    return a ^ b

def xnor_ref(a, b):
    return ~(a ^ b)

def not_ref(a):
    return ~a

def copy_ref(a):
    return a

def nor_ref(a, b):
    return ~(a + b)

def andny_ref(a, b):
    return ~a * b

def andyn_ref(a, b):
    return a * ~b

def orny_ref(a, b):
    return ~a + b

def oryn_ref(a, b):
    return a + ~b

def mux_ref(a, b, c):
    return a * b + ~a * c

def uint_min_ref(p1, p2):
    ints1 = bitarray_to_uintarray(p1)
    ints2 = bitarray_to_uintarray(p2)
    ires = numpy.minimum(ints1, ints2)
    res = uintarray_to_bitarray(ires)
    return res


def test_nand_gate(thread, key_pair):
    check_gate(thread, key_pair, 2, gate_nand, nand_ref)


def test_or_gate(thread, key_pair):
    check_gate(thread, key_pair, 2, gate_or, or_ref)


def test_and_gate(thread, key_pair):
    check_gate(thread, key_pair, 2, gate_and, and_ref)


def test_xor_gate(thread, key_pair):
    check_gate(thread, key_pair, 2, gate_xor, xor_ref)


def test_xnor_gate(thread, key_pair):
    check_gate(thread, key_pair, 2, gate_xnor, xnor_ref)


def test_not_gate(thread, key_pair):
    check_gate(thread, key_pair, 1, gate_not, not_ref)


def test_copy_gate(thread, key_pair):
    check_gate(thread, key_pair, 1, gate_copy, copy_ref)


def test_nor_gate(thread, key_pair):
    check_gate(thread, key_pair, 2, gate_nor, nor_ref)


def test_andny_gate(thread, key_pair):
    check_gate(thread, key_pair, 2, gate_andny, andny_ref)


def test_andyn_gate(thread, key_pair):
    check_gate(thread, key_pair, 2, gate_andyn, andyn_ref)


def test_orny_gate(thread, key_pair):
    check_gate(thread, key_pair, 2, gate_orny, orny_ref)


def test_oryn_gate(thread, key_pair):
    check_gate(thread, key_pair, 2, gate_oryn, oryn_ref)


def test_mux_gate(thread, key_pair):
    check_gate(thread, key_pair, 3, gate_mux, mux_ref)


def test_constant_gate(thread, key_pair):
    # Not using check_gate(), because no encryption is required.

    size = 32

    secret_key, cloud_key = key_pair
    rng = DeterministicRNG()

    params = cloud_key.params
    answer = empty_ciphertext(thread, params, (size,))

    vals = get_plaintexts(rng, 1, shape=(size,))[0]

    gate_constant(thread, cloud_key, answer, vals)
    answer_bits = decrypt(thread, secret_key, answer)
    assert (answer_bits == vals).all()


def test_uint_min(thread, key_pair):
    check_gate(thread, key_pair, 2, uint_min, uint_min_ref, shape=(4, 16))


def check_performance(
        thread, key_pair, perf_params, shape, test_function=(gate_nand, nand_ref, 2)):

    # Assuming that the time taken by the gate has the form
    #   t = size * speed + overhead
    # Then, for two results t(size1), t(size2):
    #   speed = (t(size1) - t(size2)) / (size1 - size2)
    #   overhead = (t(size1) * size2 - t(size2) * size1) / (size2 - size1)

    nufhe_func, ref_func, nargs = test_function

    if isinstance(shape, tuple):
        shape1 = shape
        shape2 = (shape[0] // 2,) + shape[1:]
        size1 = numpy.prod(shape1)
        size2 = numpy.prod(shape2)
    else:
        shape1 = shape
        shape2 = shape // 2
        size1 = shape1
        size2 = shape2

    times1 = check_gate(
        thread, key_pair, nargs, nufhe_func, ref_func,
        shape=shape1, performance_test=True, perf_params=perf_params)
    times2 = check_gate(
        thread, key_pair, nargs, nufhe_func, ref_func,
        shape=shape2, performance_test=True, perf_params=perf_params)

    mean1 = times1.mean()
    err1 = times1.std() / times1.size**0.5
    mean2 = times2.mean()
    err2 = times2.std() / times2.size**0.5

    speed_overall_mean = mean1 / size1
    speed_overall_err = err1 / size1

    speed_mean = (mean1 - mean2) / (size1 - size2)
    speed_err = abs((err1 + err2) / (size1 - size2))

    overhead_mean = (mean1 * size2 - mean2 * size1) / (size2 - size1)
    overhead_err = abs((err1 * size2 + err2 * size2) / (size2 - size1))

    return dict(
        speed_overall_mean=speed_overall_mean,
        speed_overall_err=speed_overall_err,
        speed_mean=speed_mean,
        speed_err=speed_err,
        overhead_mean=overhead_mean,
        overhead_err=overhead_err)


def check_performance_str(results):
    return (
        "Overall speed: {somean:.4f} +/- {soerr:.4f} ms/bit, " +
        "scaled: {smean:.4f} +/- {serr:.4f} ms/bit, " +
        "overhead: {omean:.4f} +/- {oerr:.4f} ms").format(
        somean=results['speed_overall_mean'] * 1e3,
        soerr=results['speed_overall_err'] * 1e3,
        smean=results['speed_mean'] * 1e3,
        serr=results['speed_err'] * 1e3,
        omean=results['overhead_mean'] * 1e3,
        oerr=results['overhead_err'] * 1e3)


@pytest.mark.perf
@pytest.mark.parametrize('test_function_name', ['NAND', 'MUX', 'uint_min'])
def test_single_kernel_bs_performance(
        thread, transform_type, single_kernel_bootstrap,
        test_function_name, heavy_performance_load):

    if not transform_supported(thread.device_params, transform_type):
        pytest.skip()

    test_function = dict(
        NAND=(gate_nand, nand_ref, 2),
        MUX=(gate_mux, mux_ref, 3),
        uint_min=(uint_min, uint_min_ref, 2),
        )[test_function_name]

    if test_function_name == 'uint_min':
        shape = (128, 32) if heavy_performance_load else (4, 16)
    else:
        shape = 4096 if heavy_performance_load else 64

    rng = DeterministicRNG()
    secret_key, cloud_key = make_key_pair(thread, rng, transform_type=transform_type)

    # TODO: instead of creating a whole key and then checking if the parameters are supported,
    # we can just create a parameter object separately.
    if (single_kernel_bootstrap
            and not single_kernel_bootstrap_supported(secret_key.params, thread.device_params)):
        pytest.skip()

    perf_params = PerformanceParameters(
        secret_key.params, single_kernel_bootstrap=single_kernel_bootstrap)
    perf_params = perf_params.for_device(thread.device_params)

    results = check_performance(
        thread, (secret_key, cloud_key), perf_params, shape=shape, test_function=test_function)
    print()
    print(check_performance_str(results))


@pytest.mark.perf
@pytest.mark.parametrize('use_constant_memory', [False, True], ids=['global_mem', 'constant_mem'])
def test_constant_mem_performance(
        thread, transform_type, single_kernel_bootstrap, heavy_performance_load,
        use_constant_memory):

    if not transform_supported(thread.device_params, transform_type):
        pytest.skip()

    # We want to test the effect of using constant memory on the bootstrap calculation.
    # A single-kernel bootstrap uses the `use_constant_memory_multi_iter` option,
    # and a multi-kernel bootstrap uses the `use_constant_memory_single_iter` option.
    kwds = dict(single_kernel_bootstrap=single_kernel_bootstrap)
    if single_kernel_bootstrap:
        kwds.update(dict(use_constant_memory_multi_iter=use_constant_memory))
    else:
        kwds.update(dict(use_constant_memory_single_iter=use_constant_memory))

    size = 4096 if heavy_performance_load else 64

    rng = DeterministicRNG()
    secret_key, cloud_key = make_key_pair(thread, rng, transform_type=transform_type)

    # TODO: instead of creating a whole key and then checking if the parameters are supported,
    # we can just create a parameter object separately.
    if (single_kernel_bootstrap
            and not single_kernel_bootstrap_supported(secret_key.params, thread.device_params)):
        pytest.skip()

    perf_params = PerformanceParameters(secret_key.params, **kwds).for_device(thread.device_params)

    results = check_performance(thread, (secret_key, cloud_key), perf_params, shape=size)
    print()
    print(check_performance_str(results))


@pytest.mark.perf
@pytest.mark.parametrize(
    'transforms_per_block', [1, 2, 3, 4], ids=['tpb=1', 'tpb=2', 'tpb=3', 'tpb=4'])
def test_transforms_per_block_performance(
        thread, transform_type, heavy_performance_load, transforms_per_block):

    if not transform_supported(thread.device_params, transform_type):
        pytest.skip()

    max_tpb = max_supported_transforms_per_block(thread.device_params, transform_type)
    if transforms_per_block > max_tpb:
        pytest.skip()

    size = 4096 if heavy_performance_load else 64

    rng = DeterministicRNG()
    secret_key, cloud_key = make_key_pair(thread, rng, transform_type=transform_type)

    perf_params = PerformanceParameters(
        secret_key.params,
        single_kernel_bootstrap=False,
        transforms_per_block=transforms_per_block).for_device(thread.device_params)

    results = check_performance(thread, (secret_key, cloud_key), perf_params, shape=size)
    print()
    print(check_performance_str(results))


@pytest.mark.perf
@pytest.mark.parametrize(
    'ntt_base_method', ['cuda_asm', 'c'], ids=['ntt_base=cuda_asm', 'ntt_base=c'])
def test_ntt_base_method_performance(
        thread, single_kernel_bootstrap, heavy_performance_load, ntt_base_method):

    if thread.api.get_id() != cuda_id() and ntt_base_method == 'cuda_asm':
        pytest.skip()

    size = 4096 if heavy_performance_load else 64

    rng = DeterministicRNG()
    secret_key, cloud_key = make_key_pair(thread, rng, transform_type='NTT')

    # TODO: instead of creating a whole key and then checking if the parameters are supported,
    # we can just create a parameter object separately.
    if (single_kernel_bootstrap
            and not single_kernel_bootstrap_supported(secret_key.params, thread.device_params)):
        pytest.skip()

    perf_params = PerformanceParameters(
        secret_key.params,
        single_kernel_bootstrap=single_kernel_bootstrap,
        ntt_base_method=ntt_base_method).for_device(thread.device_params)

    results = check_performance(thread, (secret_key, cloud_key), perf_params, shape=size)
    print()
    print(check_performance_str(results))


@pytest.mark.perf
@pytest.mark.parametrize(
    'ntt_mul_method',
    ['cuda_asm', 'c_from_asm', 'c'],
    ids=['ntt_mul=cuda_asm', 'ntt_mul=c_from_asm', 'ntt_mul=c'])
def test_ntt_mul_method_performance(
        thread, single_kernel_bootstrap, heavy_performance_load, ntt_mul_method):

    if thread.api.get_id() != cuda_id() and ntt_mul_method == 'cuda_asm':
        pytest.skip()

    size = 4096 if heavy_performance_load else 64

    rng = DeterministicRNG()
    secret_key, cloud_key = make_key_pair(thread, rng, transform_type='NTT')

    # TODO: instead of creating a whole key and then checking if the parameters are supported,
    # we can just create a parameter object separately.
    if (single_kernel_bootstrap
            and not single_kernel_bootstrap_supported(secret_key.params, thread.device_params)):
        pytest.skip()

    perf_params = PerformanceParameters(
        secret_key.params,
        single_kernel_bootstrap=single_kernel_bootstrap,
        ntt_mul_method=ntt_mul_method).for_device(thread.device_params)

    results = check_performance(thread, (secret_key, cloud_key), perf_params, shape=size)
    print()
    print(check_performance_str(results))


@pytest.mark.perf
@pytest.mark.parametrize(
    'ntt_lsh_method',
    ['cuda_asm', 'c_from_asm', 'c'],
    ids=['ntt_lsh=cuda_asm', 'ntt_lsh=c_from_asm', 'ntt_lsh=c'])
def test_ntt_lsh_method_performance(
        thread, single_kernel_bootstrap, heavy_performance_load, ntt_lsh_method):

    if thread.api.get_id() != cuda_id() and ntt_lsh_method == 'cuda_asm':
        pytest.skip()

    size = 4096 if heavy_performance_load else 64

    rng = DeterministicRNG()
    secret_key, cloud_key = make_key_pair(thread, rng, transform_type='NTT')

    # TODO: instead of creating a whole key and then checking if the parameters are supported,
    # we can just create a parameter object separately.
    if (single_kernel_bootstrap
            and not single_kernel_bootstrap_supported(secret_key.params, thread.device_params)):
        pytest.skip()

    perf_params = PerformanceParameters(
        secret_key.params,
        single_kernel_bootstrap=single_kernel_bootstrap,
        ntt_lsh_method=ntt_lsh_method).for_device(thread.device_params)

    results = check_performance(thread, (secret_key, cloud_key), perf_params, shape=size)
    print()
    print(check_performance_str(results))


def test_gate_over_view(thread, key_pair, single_kernel_bootstrap):

    secret_key, cloud_key = key_pair
    params = cloud_key.params

    if (single_kernel_bootstrap
            and not single_kernel_bootstrap_supported(params, thread.device_params)):
        pytest.skip()

    perf_params = PerformanceParameters(params, single_kernel_bootstrap=single_kernel_bootstrap)
    perf_params = perf_params.for_device(thread.device_params)

    nufhe_func = gate_nand
    reference_func = nand_ref
    num_arguments = 2

    rng = DeterministicRNG()

    shape = (5, 8)

    # FIXME: negative steps are supported as well, but the current stable PyCUDA
    # has a bug where in that case it calculates strides incorrectly.
    # It is fixed in the trunk, so we must add some negative steps here as soon as it is released.
    slices1 = (slice(3, 5), slice(1, 7, 2))
    slices2 = (slice(1, 3), slice(2, 8, 2))
    result_slices = (slice(2, 4), slice(0, 6, 2))

    plaintexts = get_plaintexts(rng, num_arguments, shape=shape)
    pt1 = plaintexts[0][slices1]
    pt2 = plaintexts[1][slices2]

    ciphertexts = [encrypt(thread, rng, secret_key, plaintext) for plaintext in plaintexts]
    ct1 = ciphertexts[0][slices1]
    ct2 = ciphertexts[1][slices2]

    reference = reference_func(pt1, pt2)

    answer = empty_ciphertext(thread, params, shape)
    answer_view = answer[result_slices]

    nufhe_func(thread, cloud_key, answer_view, ct1, ct2, perf_params=perf_params)

    answer_bits = decrypt(thread, secret_key, answer)
    answer_bits_view = answer_bits[result_slices]

    assert (answer_bits_view == reference).all()



# Copyright (C) 2018 NuCypher
#
# This file is part of nufhe.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

import numpy

from .api_low_level import empty_ciphertext
from .performance import PerformanceParameters, PerformanceParametersForDevice
from .gates import (
    gate_constant,
    gate_xnor,
    gate_mux,
    )


def _uint_to_bits(x, bitsize):
    return numpy.array([((x >> i) & 1 != 0) for i in reversed(range(bitsize))])


def _bits_to_uint(bits, dtype):
    int_answer = 0
    for i in range(bits.size):
        int_answer = int_answer | (bits[i] << (bits.size - i - 1))
    return dtype(int_answer)


def uintarray_to_bitarray(xs, itemsize=None):
    if itemsize is None:
        itemsize = xs.itemsize * 8
    assert numpy.issubdtype(xs.dtype, numpy.unsignedinteger)
    res = numpy.vstack(_uint_to_bits(x, itemsize) for x in xs.flatten())
    return res.reshape(xs.shape + (itemsize,))


def bitarray_to_uintarray(xs):

    itemsize = xs.shape[-1]
    dtype = {
        8: numpy.uint8,
        16: numpy.uint16,
        32: numpy.uint32,
        64: numpy.uint64}[itemsize]

    ints = []
    flat_xs = xs.reshape(numpy.prod(xs.shape[:-1]), xs.shape[-1])
    for j in range(flat_xs.shape[0]):
        ints.append(_bits_to_uint(flat_xs[j], dtype))
    return numpy.array(ints).reshape(xs.shape[:-1])


def uint_min(thread, cloud_key, answer, a, b, perf_params: PerformanceParametersForDevice=None):

    if perf_params is None:
        perf_params = PerformanceParameters(bk.params).for_device(thread.device_params)

    params = cloud_key.params

    itemsize = answer.shape_info.shape[-1]

    tmp1 = empty_ciphertext(thread, params, a.shape_info.shape[:-1] + (1,))
    tmp2 = empty_ciphertext(thread, params, a.shape_info.shape[:-1] + (1,))

    # initialize the carry to 0
    gate_constant(thread, cloud_key, tmp1, False)

    # Compare i-th bits in turn starting from the end (assuming big-endian order).
    # Store the result in tmp2, and use tmp1 as an accumulator.
    # Elementary full comparator gate that is used to compare the i-th bit.
    #   input: ai and bi the i-th bit of a and b
    #          lsb_carry: the result of the comparison on the lowest bits
    #   algo: if (a==b) return lsb_carry else return b
    for i in reversed(range(itemsize)):
        a_slice = a[:,i:i+1]
        b_slice = b[:,i:i+1]

        # tmp2 = (a_bit == b_bit)
        gate_xnor(thread, cloud_key, tmp2, a_slice, b_slice, perf_params=perf_params)
        # tmp1 = tmp2 ? tmp1 : a_bit
        gate_mux(thread, cloud_key, tmp1, tmp2, tmp1, a_slice, perf_params=perf_params)

    # tmp1 is the result of the comparaison: 0 if a is smaller, 1 if b is smaller
    gate_mux(thread, cloud_key, answer, tmp1, b, a, perf_params=perf_params)



# Copyright (C) 2018 NuCypher
#
# This file is part of nufhe.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

import numpy

from reikna.algorithms import PureParallel

from nufhe.tgsw import TGswParams, TGswSampleArray, TransformedTGswSampleArray
from nufhe.tlwe import TLweSampleArray
from nufhe.api_low_level import NuFHEParameters
from nufhe.numeric_functions import Torus32, Int32
from nufhe.polynomial_transform import get_transform
from nufhe.tgsw_gpu import (
    get_tgsw_polynomial_decomp_trf,
    get_tlwe_transformed_add_mul_to_trf,
    TGswTransformedExternalMul,
    TGswAddMessage,
    )
from nufhe.tgsw_cpu import (
    tgsw_polynomial_decomp_trf_reference,
    tlwe_transformed_add_mul_to_trf_reference,
    TGswTransformedExternalMulReference,
    TGswAddMessageReference,
    )
from nufhe.performance import PerformanceParameters

from utils import get_test_array, errors_allclose


def test_tgsw_polynomial_decomp_trf(thread):

    shape = (2, 3)
    params = NuFHEParameters()
    tgsw_params = params.tgsw_params
    decomp_length = tgsw_params.decomp_length
    mask_size = tgsw_params.tlwe_params.mask_size
    polynomial_degree = tgsw_params.tlwe_params.polynomial_degree

    sample = get_test_array(shape + (mask_size + 1, polynomial_degree), Torus32, (0, 1000))
    result = numpy.empty(shape + (mask_size + 1, decomp_length, polynomial_degree), dtype=Int32)

    sample_dev = thread.to_device(sample)
    result_dev = thread.empty_like(result)

    trf = get_tgsw_polynomial_decomp_trf(tgsw_params, shape)
    test = PureParallel.from_trf(trf, guiding_array='result').compile(thread)

    ref = tgsw_polynomial_decomp_trf_reference(tgsw_params, shape)

    test(result_dev, sample_dev)
    result_test = result_dev.get()

    ref(result, sample)

    assert (result == result_test).all()


def test_tlwe_transformed_add_mul_to_trf(thread):

    shape = (2, 3)
    params = NuFHEParameters(transform_type='NTT')
    perf_params = PerformanceParameters(params).for_device(thread.device_params)
    tgsw_params = params.tgsw_params

    decomp_length = tgsw_params.decomp_length
    mask_size = tgsw_params.tlwe_params.mask_size
    polynomial_degree = tgsw_params.tlwe_params.polynomial_degree

    transform_type = tgsw_params.tlwe_params.transform_type
    transform = get_transform(transform_type)
    tlength = transform.transformed_length(polynomial_degree)
    tdtype = transform.transformed_dtype()

    result_shape = shape + (mask_size + 1, tlength)
    sample_shape = shape + (mask_size + 1, decomp_length, tlength)
    bk_len = 10
    bootstrap_key_shape = (bk_len, mask_size + 1, decomp_length, mask_size + 1, tlength)
    bk_row_idx = 2

    result = numpy.empty(result_shape, tdtype)

    sample = get_test_array(sample_shape, 'ff_number')
    bootstrap_key = get_test_array(bootstrap_key_shape, 'ff_number')

    result_dev = thread.empty_like(result)
    sample_dev = thread.to_device(sample)
    bootstrap_key_dev = thread.to_device(bootstrap_key)

    trf = get_tlwe_transformed_add_mul_to_trf(tgsw_params, shape, bk_len, perf_params)
    test = PureParallel.from_trf(trf, guiding_array='result').compile(thread)
    ref = tlwe_transformed_add_mul_to_trf_reference(tgsw_params, shape, bk_len, perf_params)

    test(result_dev, sample_dev, bootstrap_key_dev, bk_row_idx)
    result_test = result_dev.get()

    ref(result, sample, bootstrap_key, bk_row_idx)

    if numpy.issubdtype(tdtype, numpy.integer):
        assert (result == result_test).all()
    else:
        assert numpy.allclose(result, result_test)


def test_tgsw_transformed_external_mul(thread):

    shape = (2, 3)
    params = NuFHEParameters(transform_type='NTT')
    perf_params = PerformanceParameters(params).for_device(thread.device_params)
    tgsw_params = params.tgsw_params

    decomp_length = tgsw_params.decomp_length
    mask_size = tgsw_params.tlwe_params.mask_size
    polynomial_degree = tgsw_params.tlwe_params.polynomial_degree

    transform_type = tgsw_params.tlwe_params.transform_type
    transform = get_transform(transform_type)
    tlength = transform.transformed_length(polynomial_degree)
    tdtype = transform.transformed_dtype()

    accum_shape = shape + (mask_size + 1, polynomial_degree)
    bk_len = 10
    bootstrap_key_shape = (bk_len, mask_size + 1, decomp_length, mask_size + 1, tlength)
    bk_row_idx = 2

    bootstrap_key = get_test_array(bootstrap_key_shape, 'ff_number')
    accum = get_test_array(accum_shape, Torus32, (-1000, 1000))

    bootstrap_key_dev = thread.to_device(bootstrap_key)
    accum_dev = thread.to_device(accum)
    thread.synchronize()

    test = TGswTransformedExternalMul(tgsw_params, shape, bk_len, perf_params).compile(thread)
    ref = TGswTransformedExternalMulReference(tgsw_params, shape, bk_len, perf_params)

    test(accum_dev, bootstrap_key_dev, bk_row_idx)
    accum_test = accum_dev.get()

    ref(accum, bootstrap_key, bk_row_idx)

    assert (accum == accum_test).all()


def test_tgsw_add_message(thread):

    params = NuFHEParameters()
    tgsw_params = params.tgsw_params

    decomp_length = tgsw_params.decomp_length
    mask_size = tgsw_params.tlwe_params.mask_size
    polynomial_degree = tgsw_params.tlwe_params.polynomial_degree

    shape = (3, 5)

    result_a = get_test_array(
        shape + (mask_size + 1, decomp_length, mask_size + 1, polynomial_degree), Torus32)
    messages = get_test_array(shape, Torus32)

    result_a_dev = thread.to_device(result_a)
    messages_dev = thread.to_device(messages)
    thread.synchronize()

    test = TGswAddMessage(tgsw_params, shape).compile(thread)
    ref = TGswAddMessageReference(tgsw_params, shape)

    test(result_a_dev, messages_dev)
    ref(result_a, messages)

    result_a_test = result_a_dev.get()

    assert (result_a_test == result_a).all()



# Copyright (C) 2018 NuCypher
#
# This file is part of nufhe.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

import numpy
import pytest

from reikna.cluda import cuda_api, ocl_api, get_api, supported_api_ids, find_devices

from nufhe import make_key_pair, DeterministicRNG, Context
from nufhe.computation_cache import clear_computation_cache


def pytest_addoption(parser):
    api_ids = supported_api_ids()

    parser.addoption("--api", action="store",
        help="Backend API. If 'supported' is chosen, will run on all available ones.",
        default="supported", choices=api_ids + ["supported"])
    parser.addoption("--device-include-mask", action="append",
        help="Run tests on matching devices only",
        default=[])
    parser.addoption("--include-duplicate-devices", action="store_true",
        help="Run tests on all available devices and not only on uniquely named ones.",
        default=False)
    parser.addoption("--heavy-performance-load", action="store_true",
        help=(
            "Use large data sizes and numbers of iterations for performance tests. "
            "Recommended for high-tier videocards."),
        default=False)
    parser.addoption("--transform", action="store",
        help=(
            "The type of polynomial transform to use for tests "
            "that can use different transform types."),
        default="all", choices=["NTT", "FFT", "all"])


def pytest_generate_tests(metafunc):

    config = metafunc.config

    if 'thread' in metafunc.fixturenames:

        api_opt = config.option.api
        api_ids = supported_api_ids() if api_opt == 'supported' else [api_opt]

        vals = []
        ids = []
        for api_id in api_ids:
            api = get_api(api_id)
            devices = find_devices(
                api,
                include_devices=config.option.device_include_mask,
                include_duplicate_devices=config.option.include_duplicate_devices)
            for pnum in sorted(devices.keys()):
                dnums = sorted(devices[pnum])
                for dnum in dnums:
                    vals.append((api_id, pnum, dnum))
                    ids.append("{api_id}:{pnum}:{dnum}".format(api_id=api_id, pnum=pnum, dnum=dnum))

        if len(vals) == 0:
            raise RuntimeError(
                "Neither PyCUDA nor PyOpenCL could find any suitable devices. "
                "Check your system configuration.")

        metafunc.parametrize("thread", vals, ids=ids, indirect=True)

    if 'transform_type' in metafunc.fixturenames:

        if config.option.transform == 'all':
            vals = ['NTT', 'FFT']
        else:
            vals = [config.option.transform]

        metafunc.parametrize("transform_type", vals)


@pytest.fixture(scope='session')
def thread(request):

    api_id, pnum, dnum = request.param

    api = get_api(api_id)

    platform = api.get_platforms()[pnum]
    device = platform.get_devices()[dnum]

    thread = api.Thread(device)
    yield thread

    # Computations may retain references to the Thread objects,
    # so we need to clear the cache first so that the thread could be destroyed
    # as it goes out of scope.
    # CUDA is sensitive to the exact timing of the destruction
    # because of the stateful nature of its API.
    clear_computation_cache(thread)


@pytest.fixture(scope='session')
def heavy_performance_load(request):
    return request.config.option.heavy_performance_load


@pytest.fixture(scope='session')
def key_pair(thread):
    rng = DeterministicRNG()
    secret_key, cloud_key = make_key_pair(thread, rng)
    return secret_key, cloud_key


@pytest.fixture(scope='session')
def context(thread):
    return Context(thread=thread)


@pytest.fixture(scope='session')
def context_and_key_pair(context):
    secret_key, cloud_key = context.make_key_pair()
    return context, secret_key, cloud_key



# Copyright (C) 2018 NuCypher
#
# This file is part of nufhe.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

from setuptools import setup

setup(
    name='nufhe',
    version='0.0.3',
    description='A GPU implementation of fully homomorphic encryption on torus',
    url='http://github.com/nucypher/nufhe',
    author='Bogdan Opanchuk',
    author_email='bogdan@nucypher.com',
    license='GPLv3',
    packages=[
        'nufhe',
        'nufhe/transform',
        ],
    package_data={
        'nufhe': ['*.mako'],
        'nufhe/transform': ['*.mako'],
        },
    install_requires=['numpy', 'reikna>=0.7.4'],
    python_requires='>=3.5',
    extras_require={
        'dev': ['pytest', 'pytest-cov', 'sphinx', 'sphinx_autodoc_typehints'],
        'pyopencl': ['pyopencl>=2018.1.1'],
        'pycuda': ['pycuda>=2018.1.1'],
        },
    zip_safe=True)
